#
# OpenLDAP directory server instance class
#
# Copyright (c) 2015 Red Hat, Inc.
# Author: Nikolai Kondrashov <Nikolai.Kondrashov@redhat.com>
# Author: Lukas Slebodnik <lslebodn@redhat.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import hashlib
import base64
import time
import ldap
import os
import errno
import signal
import shutil
import subprocess
from util import unindent, first_dir
from ds import DS

try:
    from urllib import quote as url_quote
except ImportError:
    from urllib.parse import quote as url_quote


def hash_password(password):
    """Generate userPassword value for a password."""
    salt = os.urandom(4)
    hash = hashlib.sha1(password.encode('utf-8'))
    hash.update(salt)
    hash_base64 = base64.standard_b64encode(hash.digest() + salt)
    return "{SSHA}" + hash_base64.decode('utf-8')


class DSOpenLDAP(DS):
    """OpenLDAP directory server instance."""

    def __init__(self, dir, port, base_dn, admin_rdn, admin_pw):
        """
            Initialize the instance.

            Arguments:
            dir         Path to the root of the filesystem hierarchy to create
                        the instance under.
            port        TCP port on localhost to bind the server to.
            base_dn     Base DN.
            admin_rdn   Administrator DN, relative to BASE_DN.
            admin_pw    Administrator password.
        """
        DS.__init__(self, dir, port, base_dn, admin_rdn, admin_pw)
        self.run_dir = self.dir + "/var/run/ldap"
        self.pid_path = self.run_dir + "/slapd.pid"
        self.conf_dir = self.dir + "/etc/ldap"
        self.conf_slapd_d_dir = self.conf_dir + "/slapd.d"
        self.data_dir = self.dir + "/var/lib/ldap"

    def _setup_config(self):
        """Setup the instance initial configuration."""
        dist_lib_dir = first_dir("/usr/lib64/openldap",
                                 "/usr/lib/openldap",
                                 "/usr/lib/ldap")
        dist_conf_dir = first_dir("/etc/ldap",
                                  "/etc/openldap")
        args_file = self.run_dir + "/slapd.args"
        admin_pw_hash = hash_password(self.admin_pw)
        uid = os.geteuid()
        gid = os.getegid()

        #
        # Add configuration
        #
        config = unindent("""
            dn: cn=config
            objectClass: olcGlobal
            cn: config
            olcPidFile: {self.pid_path}
            olcArgsFile: {args_file}
            # Read slapd.conf(5) for possible values
            olcLogLevel: none

            # Frontend settings
            dn: olcDatabase={{-1}}frontend,cn=config
            objectClass: olcDatabaseConfig
            objectClass: olcFrontendConfig
            olcDatabase: {{-1}}frontend
            # The maximum number of entries that is returned for
            # a search operation
            olcSizeLimit: 500
            # Allow unlimited access to local connection from the local root
            olcAccess: {{0}}to * by dn.exact=gidNumber={gid}+uidNumber={uid},
             cn=peercred,cn=external,cn=auth manage by * break
            # Allow unauthenticated read access for schema and
            # base DN autodiscovery
            olcAccess: {{1}}to dn.exact="" by * read
            olcAccess: {{2}}to dn.base="cn=Subschema" by * read

            # Config db settings
            dn: olcDatabase=config,cn=config
            objectClass: olcDatabaseConfig
            olcDatabase: config
            # Allow unlimited access to local connection from the local root
            olcAccess: to * by dn.exact=gidNumber={gid}+uidNumber={uid},
             cn=peercred,cn=external,cn=auth manage by * break
            olcRootDN: {self.admin_rdn},cn=config
            olcRootPW: {admin_pw_hash}

            # Load schemas
            dn: cn=schema,cn=config
            objectClass: olcSchemaConfig
            cn: schema

            include: file://{dist_conf_dir}/schema/core.ldif
            include: file://{dist_conf_dir}/schema/cosine.ldif
            include: file://{dist_conf_dir}/schema/nis.ldif
            include: file://{dist_conf_dir}/schema/inetorgperson.ldif

            # Load module
            dn: cn=module{{0}},cn=config
            objectClass: olcModuleList
            cn: module{{0}}
            olcModulePath: {dist_lib_dir}
            olcModuleLoad: back_mdb

            # Set defaults for the backend
            dn: olcBackend=mdb,cn=config
            objectClass: olcBackendConfig
            olcBackend: mdb

            # The database definition.
            dn: olcDatabase=mdb,cn=config
            objectClass: olcDatabaseConfig
            objectClass: olcMdbConfig
            olcDatabase: mdb
            olcDbCheckpoint: 512 30
            olcLastMod: TRUE
            olcSuffix: {self.base_dn}
            olcDbDirectory: {self.data_dir}
            olcRootDN: {self.admin_dn}
            olcRootPW: {admin_pw_hash}
            olcDbIndex: objectClass eq
            olcDbIndex: cn,uid eq
            olcDbIndex: uidNumber,gidNumber eq
            olcDbIndex: member,memberUid eq
            olcAccess: to attrs=userPassword,shadowLastChange
              by self write
              by anonymous auth
              by * none
            olcAccess: to dn.base="" by * read
            olcAccess: to *
              by * read
        """).format(**locals())

        slapadd = subprocess.Popen(
            ["slapadd", "-F", self.conf_slapd_d_dir, "-b", "cn=config"],
            stdin=subprocess.PIPE, close_fds=True
        )
        slapadd.communicate(config.encode('utf-8'))
        if slapadd.returncode != 0:
            raise Exception("Failed to add configuration with slapadd")

        #
        # Add database config (example from distribution)
        #
        db_config = unindent("""
            # One 0.25 GB cache
            set_cachesize 0 268435456 1

            # Transaction Log settings
            set_lg_regionmax 262144
            set_lg_bsize 2097152
        """)
        db_config_file = open(self.data_dir + "/DB_CONFIG", "w")
        db_config_file.write(db_config)
        db_config_file.close()

        # Import ad schema
        subprocess.check_call(
            ["slapadd", "-F", self.conf_slapd_d_dir, "-b", "cn=config",
             "-l", "data/ssh_schema.ldif"],
        )

        # Import sudo schema
        subprocess.check_call(
            ["slapadd", "-F", self.conf_slapd_d_dir, "-b", "cn=config",
             "-l", "data/sudo_schema.ldif"],
        )

        # Import cert schema
        subprocess.check_call(
            ["slapadd", "-F", self.conf_slapd_d_dir, "-b", "cn=config",
             "-l", "data/cert_schema.ldif"],
        )

    def _start_daemon(self):
        """Start the instance."""
        if subprocess.call(["slapd", "-F", self.conf_slapd_d_dir,
                            "-h", self.url_list]) != 0:
            raise Exception("Failed to start slapd")

        #
        # Wait until it is available
        #
        attempt = 0
        while True:
            try:
                ldap_conn = ldap.initialize(self.ldapi_url)
                ldap_conn.simple_bind_s(self.admin_rdn + ",cn=config",
                                        self.admin_pw)
                ldap_conn.unbind_s()
                ldap_conn = ldap.initialize(self.ldap_url)
                ldap_conn.simple_bind_s(self.admin_dn, self.admin_pw)
                ldap_conn.unbind_s()
                break
            except ldap.SERVER_DOWN:
                pass
            attempt = attempt + 1
            if attempt > 30:
                raise Exception("Failed to start slapd")
            time.sleep(1)

    def setup(self):
        """Setup the instance."""
        ldapi_socket = self.run_dir + "/ldapi"
        self.ldapi_url = "ldapi://" + url_quote(ldapi_socket, "")
        self.url_list = self.ldapi_url + " " + self.ldap_url

        os.makedirs(self.conf_slapd_d_dir)
        os.makedirs(self.run_dir)
        os.makedirs(self.data_dir)

        #
        # Setup initial configuration
        #
        self._setup_config()

        self._start_daemon()

        #
        # Relax requirement of member attribute presence in groupOfNames
        #
        modlist = [
            (ldap.MOD_DELETE, "olcObjectClasses",
             b"{7}( 2.5.6.9 NAME 'groupOfNames' "
             b"DESC 'RFC2256: a group of names (DNs)' SUP top "
             b"STRUCTURAL MUST ( member $ cn ) MAY ( businessCategory $ "
             b"seeAlso $ owner $ ou $ o $ description ) )"),
            (ldap.MOD_ADD, "olcObjectClasses",
             b"{7}( 2.5.6.9 NAME 'groupOfNames' "
             b"DESC 'RFC2256: a group of names (DNs)' SUP top "
             b"STRUCTURAL MUST ( cn ) MAY ( member $ businessCategory $ "
             b"seeAlso $ owner $ ou $ o $ description ) )"),
        ]
        ldap_conn = ldap.initialize(self.ldapi_url)
        ldap_conn.simple_bind_s(self.admin_rdn + ",cn=config", self.admin_pw)
        ldap_conn.modify_s("cn={0}core,cn=schema,cn=config", modlist)
        ldap_conn.unbind_s()

        #
        # Add data
        #
        ldap_conn = ldap.initialize(self.ldap_url)
        ldap_conn.simple_bind_s(self.admin_dn, self.admin_pw)
        ldap_conn.add_s(self.base_dn, [
            ("objectClass", [b"dcObject", b"organization"]),
            ("o", b"Example Company"),
        ])
        ldap_conn.add_s("cn=Manager," + self.base_dn, [
            ("objectClass", b"organizationalRole"),
        ])
        for ou in ("Users", "Groups", "Netgroups", "Services", "Policies",
                   "Hosts", "Networks"):
            ldap_conn.add_s("ou=" + ou + "," + self.base_dn, [
                ("objectClass", [b"top", b"organizationalUnit"]),
            ])
        ldap_conn.add_s("ou=sudoers," + self.base_dn, [
            ("objectClass", [b"top", b"organizationalUnit"]),
        ])
        ldap_conn.add_s("cn=testrule,ou=sudoers," + self.base_dn, [
            ("objectClass", [b"top", b"sudoRole"]),
            ("sudoUser", [b"tuser"]),
        ])
        ldap_conn.unbind_s()

    def _stop_daemon(self):
        """Stop the instance."""
        # Wait for slapd to stop
        try:
            pid_file = open(self.pid_path, "r")
            try:
                os.kill(int(pid_file.read()), signal.SIGTERM)
            finally:
                pid_file.close()
            attempt = 0
            while os.path.isfile(self.pid_path):
                attempt = attempt + 1
                if attempt > 30:
                    raise Exception("Failed to stop slapd")
                time.sleep(1)
        except IOError as e:
            if e.errno != errno.ENOENT:
                raise

    def teardown(self):
        """Teardown the instance."""
        self._stop_daemon()

        for path in (self.conf_slapd_d_dir, self.run_dir, self.data_dir):
            shutil.rmtree(path, True)


class FakeAD(DSOpenLDAP):
    """Fake Active Directory based on OpenLDAP directory server."""

    def _setup_config(self):
        """Setup the instance initial configuration."""

        # Import ad schema
        subprocess.check_call(
            ["slapadd", "-F", self.conf_slapd_d_dir, "-b", "cn=config",
             "-l", "data/ad_schema.ldif"],
        )

    def setup(self):
        """Setup the instance."""
        ldapi_socket = self.run_dir + "/ldapi"
        self.ldapi_url = "ldapi://" + url_quote(ldapi_socket, "")
        self.url_list = self.ldapi_url + " " + self.ldap_url

        os.makedirs(self.conf_slapd_d_dir)
        os.makedirs(self.run_dir)
        os.makedirs(self.data_dir)

        super(FakeAD, self)._setup_config()
        self._setup_config()

        # Start the daemon
        super(FakeAD, self)._start_daemon()

        # Relax requirement of surname attribute presence in person
        modlist = [
            (ldap.MOD_DELETE, "olcObjectClasses",
             b"{4}( 2.5.6.6 NAME 'person' DESC 'RFC2256: a person' SUP top "
             b"STRUCTURAL MUST ( sn $ cn ) MAY ( userPassword $ "
             b"telephoneNumber $ seeAlso $ description ) )"),
            (ldap.MOD_ADD, "olcObjectClasses",
             b"{4}( 2.5.6.6 NAME 'person' DESC 'RFC2256: a person' SUP top "
             b"STRUCTURAL MUST ( cn ) MAY ( sn $ userPassword $ "
             b"telephoneNumber $ seeAlso $ description ) )"),
        ]
        ldap_conn = ldap.initialize(self.ldapi_url)
        ldap_conn.simple_bind_s(self.admin_rdn + ",cn=config", self.admin_pw)
        ldap_conn.modify_s("cn={0}core,cn=schema,cn=config", modlist)
        ldap_conn.unbind_s()

        # restart daemon for reloading schema
        super(FakeAD, self)._stop_daemon()
        super(FakeAD, self)._start_daemon()

        # Add data
        ldap_conn = ldap.initialize(self.ldap_url)
        ldap_conn.simple_bind_s(self.admin_dn, self.admin_pw)
        ldap_conn.add_s(self.base_dn, [
            ("objectClass", [b"dcObject", b"organization"]),
            ("o", b"Example Company"),
        ])
        ldap_conn.add_s("cn=Manager," + self.base_dn, [
            ("objectClass", b"organizationalRole"),
        ])
        for ou in ("Users", "Groups", "Netgroups", "Services", "Policies"):
            ldap_conn.add_s("ou=" + ou + "," + self.base_dn, [
                ("objectClass", [b"top", b"organizationalUnit"]),
            ])
        ldap_conn.unbind_s()

        # import data from real AD
        subprocess.check_call(
            ["ldapadd", "-x", "-w", self.admin_pw, "-D",
             self.admin_dn, "-H", self.ldap_url,
             "-f", "data/ad_data.ldif"],
        )

    def teardown(self):
        """Teardown the instance."""
        super(FakeAD, self).teardown()



#
# Netgroup integration test
#
# Copyright (c) 2016 Red Hat, Inc.
# Author: Petr Cech <pcech@redhat.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import stat
import signal
import subprocess
import time
import ldap
import ldap.modlist
import pytest

import config
import ds_openldap
import ldap_ent
from util import unindent
from sssd_nss import NssReturnCode
from sssd_netgroup import get_sssd_netgroups

LDAP_BASE_DN = "dc=example,dc=com"


@pytest.fixture(scope="module")
def ds_inst(request):
    """LDAP server instance fixture"""
    ds_inst = ds_openldap.DSOpenLDAP(
        config.PREFIX, 10389, LDAP_BASE_DN,
        "cn=admin", "Secret123"
    )

    try:
        ds_inst.setup()
    except Exception:
        ds_inst.teardown()
        raise
    request.addfinalizer(ds_inst.teardown)
    return ds_inst


@pytest.fixture(scope="module")
def ldap_conn(request, ds_inst):
    """LDAP server connection fixture"""
    ldap_conn = ds_inst.bind()
    ldap_conn.ds_inst = ds_inst
    request.addfinalizer(ldap_conn.unbind_s)
    return ldap_conn


def create_ldap_entries(ldap_conn, ent_list=None):
    """Add LDAP entries from ent_list"""
    if ent_list is not None:
        for entry in ent_list:
            ldap_conn.add_s(entry[0], entry[1])


def cleanup_ldap_entries(ldap_conn, ent_list=None):
    """Remove LDAP entries added by create_ldap_entries"""
    if ent_list is None:
        for ou in ("Users", "Groups", "Netgroups", "Services", "Policies"):
            for entry in ldap_conn.search_s(f"ou={ou},"
                                            f"{ldap_conn.ds_inst.base_dn}",
                                            ldap.SCOPE_ONELEVEL,
                                            attrlist=[]):
                ldap_conn.delete_s(entry[0])
    else:
        for entry in ent_list:
            ldap_conn.delete_s(entry[0])


def create_ldap_cleanup(request, ldap_conn, ent_list=None):
    """Add teardown for removing all user/group LDAP entries"""
    request.addfinalizer(lambda: cleanup_ldap_entries(ldap_conn, ent_list))


def create_ldap_fixture(request, ldap_conn, ent_list=None):
    """Add LDAP entries and add teardown for removing them"""
    create_ldap_entries(ldap_conn, ent_list)
    create_ldap_cleanup(request, ldap_conn, ent_list)


SCHEMA_RFC2307_BIS = "rfc2307bis"


def format_basic_conf(ldap_conn, schema):
    """Format a basic SSSD configuration"""
    schema_conf = "ldap_schema         = " + schema + "\n"
    schema_conf += "ldap_group_object_class = groupOfNames\n"
    return unindent("""\
        [sssd]
        domains             = LDAP
        services            = nss
        disable_netlink     = true

        [nss]

        [domain/LDAP]
        {schema_conf}
        id_provider         = ldap
        auth_provider       = ldap
        ldap_uri            = {ldap_conn.ds_inst.ldap_url}
        ldap_search_base    = {ldap_conn.ds_inst.base_dn}
        ldap_netgroup_search_base = ou=Netgroups,{ldap_conn.ds_inst.base_dn}
        ldap_id_use_start_tls = false
    """).format(**locals())


def create_conf_file(contents):
    """Create sssd.conf with specified contents"""
    conf = open(config.CONF_PATH, "w")
    conf.write(contents)
    conf.close()
    os.chmod(config.CONF_PATH, stat.S_IRUSR | stat.S_IWUSR)


def cleanup_conf_file():
    """Remove sssd.conf, if it exists"""
    if os.path.lexists(config.CONF_PATH):
        os.unlink(config.CONF_PATH)


def create_conf_cleanup(request):
    """Add teardown for removing sssd.conf"""
    request.addfinalizer(cleanup_conf_file)


def create_conf_fixture(request, contents):
    """
    Create sssd.conf with specified contents and add teardown for removing it
    """
    create_conf_file(contents)
    create_conf_cleanup(request)


def create_sssd_process():
    """Start the SSSD process"""
    if subprocess.call(["sssd", "-D", "--logger=files"]) != 0:
        raise Exception("sssd start failed")


def get_sssd_pid():
    pid_file = open(config.PIDFILE_PATH, "r")
    pid = int(pid_file.read())
    return pid


def cleanup_sssd_process():
    """Stop the SSSD process and remove its state"""
    try:
        pid = get_sssd_pid()
        os.kill(pid, signal.SIGTERM)
        while True:
            try:
                os.kill(pid, signal.SIGCONT)
            except OSError:
                break
            time.sleep(1)
    except OSError:
        pass
    for path in os.listdir(config.DB_PATH):
        os.unlink(config.DB_PATH + "/" + path)
    for path in os.listdir(config.MCACHE_PATH):
        os.unlink(config.MCACHE_PATH + "/" + path)


def create_sssd_cleanup(request):
    """Add teardown for stopping SSSD and removing its state"""
    request.addfinalizer(cleanup_sssd_process)


def simulate_offline():
    pid = get_sssd_pid()
    os.kill(pid, signal.SIGUSR1)


def create_sssd_fixture(request):
    """Start SSSD and add teardown for stopping it and removing its state"""
    create_sssd_process()
    create_sssd_cleanup(request)


@pytest.fixture
def add_empty_netgroup(request, ldap_conn):
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)

    ent_list.add_netgroup("empty_netgroup")

    create_ldap_fixture(request, ldap_conn, ent_list)
    conf = format_basic_conf(ldap_conn, SCHEMA_RFC2307_BIS)
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)
    return None


def test_add_empty_netgroup(add_empty_netgroup):
    """
    Adding empty netgroup.
    """

    res, _, netgroups = get_sssd_netgroups("empty_netgroup")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == []


@pytest.fixture
def add_tripled_netgroup(request, ldap_conn):
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)

    ent_list.add_netgroup("tripled_netgroup", ["(host,user,domain)"])

    ent_list.add_netgroup("adv_tripled_netgroup", ["(host1,user1,domain1)",
                                                   "(host2,user2,domain2)"])

    ent_list.add_netgroup("tripled_netgroup_no_domain", ["(host,user,)"])

    ent_list.add_netgroup("tripled_netgroup_no_user", ["(host,,domain)"])

    ent_list.add_netgroup("tripled_netgroup_no_host", ["(,user,domain)"])

    ent_list.add_netgroup("tripled_netgroup_none", ["(,,)"])

    create_ldap_fixture(request, ldap_conn, ent_list)
    conf = format_basic_conf(ldap_conn, SCHEMA_RFC2307_BIS)
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)
    return None


def test_add_tripled_netgroup(add_tripled_netgroup):
    """
    Adding netgroup with triplet.
    """

    res, _, netgrps = get_sssd_netgroups("tripled_netgroup")
    assert res == NssReturnCode.SUCCESS
    assert netgrps == [("host", "user", "domain")]

    res, _, netgrps = get_sssd_netgroups("adv_tripled_netgroup")
    assert res == NssReturnCode.SUCCESS
    assert sorted(netgrps) == sorted([("host1", "user1", "domain1"),
                                      ("host2", "user2", "domain2")])

    res, _, netgrps = get_sssd_netgroups("tripled_netgroup_no_domain")
    assert res == NssReturnCode.SUCCESS
    assert netgrps == [("host", "user", "")]

    res, _, netgrps = get_sssd_netgroups("tripled_netgroup_no_user")
    assert res == NssReturnCode.SUCCESS
    assert netgrps == [("host", "", "domain")]

    res, _, netgrps = get_sssd_netgroups("tripled_netgroup_no_host")
    assert res == NssReturnCode.SUCCESS
    assert netgrps == [("", "user", "domain")]

    res, _, netgrps = get_sssd_netgroups("tripled_netgroup_none")
    assert res == NssReturnCode.SUCCESS
    assert netgrps == [("", "", "")]


@pytest.fixture
def add_mixed_netgroup(request, ldap_conn):
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)

    ent_list.add_netgroup("mixed_netgroup1")
    ent_list.add_netgroup("mixed_netgroup2", members=["mixed_netgroup1"])

    ent_list.add_netgroup("mixed_netgroup3", ["(host1,user1,domain1)"])
    ent_list.add_netgroup("mixed_netgroup4",
                          ["(host2,user2,domain2)", "(host3,user3,domain3)"])

    ent_list.add_netgroup("mixed_netgroup5",
                          ["(host4,user4,domain4)"],
                          ["mixed_netgroup1"])
    ent_list.add_netgroup("mixed_netgroup6",
                          ["(host5,user5,domain5)"],
                          ["mixed_netgroup2"])

    ent_list.add_netgroup("mixed_netgroup7", members=["mixed_netgroup3"])
    ent_list.add_netgroup("mixed_netgroup8",
                          members=["mixed_netgroup3", "mixed_netgroup4"])

    ent_list.add_netgroup("mixed_netgroup9",
                          ["(host6,user6,domain6)"],
                          ["mixed_netgroup3", "mixed_netgroup4"])

    create_ldap_fixture(request, ldap_conn, ent_list)
    conf = format_basic_conf(ldap_conn, SCHEMA_RFC2307_BIS)
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)
    return None


def test_add_mixed_netgroup(add_mixed_netgroup):
    """
    Adding many netgroups of different type.
    """

    res, _, netgroups = get_sssd_netgroups("mixed_netgroup1")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == []

    res, _, netgroups = get_sssd_netgroups("mixed_netgroup2")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == []

    res, _, netgroups = get_sssd_netgroups("mixed_netgroup3")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == [("host1", "user1", "domain1")]

    res, _, netgroups = get_sssd_netgroups("mixed_netgroup4")
    assert res == NssReturnCode.SUCCESS
    assert sorted(netgroups) == sorted([("host2", "user2", "domain2"),
                                        ("host3", "user3", "domain3")])

    res, _, netgroups = get_sssd_netgroups("mixed_netgroup5")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == [("host4", "user4", "domain4")]

    res, _, netgroups = get_sssd_netgroups("mixed_netgroup6")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == [("host5", "user5", "domain5")]

    res, _, netgroups = get_sssd_netgroups("mixed_netgroup7")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == [("host1", "user1", "domain1")]

    res, _, netgroups = get_sssd_netgroups("mixed_netgroup8")
    assert res == NssReturnCode.SUCCESS
    assert sorted(netgroups) == sorted([("host1", "user1", "domain1"),
                                        ("host2", "user2", "domain2"),
                                        ("host3", "user3", "domain3")])

    res, _, netgroups = get_sssd_netgroups("mixed_netgroup9")
    assert res == NssReturnCode.SUCCESS
    assert sorted(netgroups) == sorted([("host1", "user1", "domain1"),
                                        ("host2", "user2", "domain2"),
                                        ("host3", "user3", "domain3"),
                                        ("host6", "user6", "domain6")])


@pytest.fixture
def remove_step_by_step(request, ldap_conn):
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)

    ent_list.add_netgroup("rm_empty_netgroup1", ["(host1,user1,domain1)"])
    ent_list.add_netgroup("rm_empty_netgroup2",
                          ["(host2,user2,domain2)"],
                          ["rm_empty_netgroup1"])

    create_ldap_fixture(request, ldap_conn, ent_list)
    conf = format_basic_conf(ldap_conn, SCHEMA_RFC2307_BIS)
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)
    return ent_list


def test_remove_step_by_step(remove_step_by_step, ldap_conn):
    """
    Removing netgroups step by step.
    """

    ent_list = remove_step_by_step

    res, _, netgroups = get_sssd_netgroups("rm_empty_netgroup1")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == [('host1', 'user1', 'domain1')]

    res, _, netgroups = get_sssd_netgroups("rm_empty_netgroup2")
    assert res == NssReturnCode.SUCCESS
    assert sorted(netgroups) == sorted([('host1', 'user1', 'domain1'),
                                        ('host2', 'user2', 'domain2')])

    # removing of rm_empty_netgroup1
    ldap_conn.delete_s(ent_list[0][0])
    ent_list.remove(ent_list[0])

    if subprocess.call(["sss_cache", "-N"]) != 0:
        raise Exception("sssd_cache failed")

    res, _, netgroups = get_sssd_netgroups("rm_empty_netgroup1")
    assert res == NssReturnCode.NOTFOUND
    assert netgroups == []

    res, _, netgroups = get_sssd_netgroups("rm_empty_netgroup2")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == [('host2', 'user2', 'domain2')]

    # removing of rm_empty_netgroup2
    ldap_conn.delete_s(ent_list[0][0])
    ent_list.remove(ent_list[0])

    if subprocess.call(["sss_cache", "-N"]) != 0:
        raise Exception("sssd_cache failed")

    res, _, netgroups = get_sssd_netgroups("rm_empty_netgroup1")
    assert res == NssReturnCode.NOTFOUND
    assert netgroups == []

    res, _, netgroups = get_sssd_netgroups("rm_empty_netgroup2")
    assert res == NssReturnCode.NOTFOUND
    assert netgroups == []


@pytest.fixture
def removing_nested_netgroups(request, ldap_conn):
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)

    ent_list.add_netgroup("t2841_netgroup1", ["(host1,user1,domain1)"])
    ent_list.add_netgroup("t2841_netgroup2", ["(host2,user2,domain2)"])
    ent_list.add_netgroup("t2841_netgroup3",
                          members=["t2841_netgroup1", "t2841_netgroup2"])

    create_ldap_fixture(request, ldap_conn, ent_list)
    conf = format_basic_conf(ldap_conn, SCHEMA_RFC2307_BIS)
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)
    return None


def test_removing_nested_netgroups(removing_nested_netgroups, ldap_conn):
    """
    Regression test for ticket 2841.
    https://fedorahosted.org/sssd/ticket/2841
    """

    netgrp_dn = 'cn=t2841_netgroup3,ou=Netgroups,' + ldap_conn.ds_inst.base_dn

    res, _, netgroups = get_sssd_netgroups("t2841_netgroup1")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == [('host1', 'user1', 'domain1')]

    res, _, netgroups = get_sssd_netgroups("t2841_netgroup2")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == [('host2', 'user2', 'domain2')]

    res, _, netgroups = get_sssd_netgroups("t2841_netgroup3")
    assert res == NssReturnCode.SUCCESS
    assert sorted(netgroups) == sorted([('host1', 'user1', 'domain1'),
                                        ('host2', 'user2', 'domain2')])

    # removing of t2841_netgroup1 from t2841_netgroup3
    old = {'memberNisNetgroup': [b"t2841_netgroup1", b"t2841_netgroup2"]}
    new = {'memberNisNetgroup': [b"t2841_netgroup2"]}

    ldif = ldap.modlist.modifyModlist(old, new)
    ldap_conn.modify_s(netgrp_dn, ldif)

    if subprocess.call(["sss_cache", "-N"]) != 0:
        raise Exception("sssd_cache failed")

    res, _, netgroups = get_sssd_netgroups("t2841_netgroup1")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == [('host1', 'user1', 'domain1')]

    res, _, netgroups = get_sssd_netgroups("t2841_netgroup2")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == [('host2', 'user2', 'domain2')]

    res, _, netgroups = get_sssd_netgroups("t2841_netgroup3")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == [('host2', 'user2', 'domain2')]

    # removing of t2841_netgroup2 from t2841_netgroup3
    old = {'memberNisNetgroup': [b"t2841_netgroup2"]}
    new = {'memberNisNetgroup': []}

    ldif = ldap.modlist.modifyModlist(old, new)
    ldap_conn.modify_s(netgrp_dn, ldif)

    if subprocess.call(["sss_cache", "-N"]) != 0:
        raise Exception("sssd_cache failed")

    res, _, netgroups = get_sssd_netgroups("t2841_netgroup1")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == [('host1', 'user1', 'domain1')]

    res, _, netgroups = get_sssd_netgroups("t2841_netgroup2")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == [('host2', 'user2', 'domain2')]

    res, _, netgroups = get_sssd_netgroups("t2841_netgroup3")
    assert res == NssReturnCode.SUCCESS
    assert netgroups == []


def test_offline_netgroups(add_tripled_netgroup):
    res, _, netgrps = get_sssd_netgroups("tripled_netgroup")
    assert res == NssReturnCode.SUCCESS
    assert netgrps == [("host", "user", "domain")]

    subprocess.check_call(["sss_cache", "-N"])

    simulate_offline()

    res, _, netgrps = get_sssd_netgroups("tripled_netgroup")
    assert res == NssReturnCode.SUCCESS
    assert netgrps == [("host", "user", "domain")]


@pytest.fixture
def add_thread_test_netgroup(request, ldap_conn):
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)

    triple_list = []
    for i in range(1, 999):
        triple_list.append("(host1,user" + str(i) + ",domain1)")
    ent_list.add_netgroup("ng1", triple_list)

    triple_list = []
    for i in range(1, 999):
        triple_list.append("(host2,user" + str(i) + ",domain2)")
    ent_list.add_netgroup("ng2", triple_list)

    create_ldap_fixture(request, ldap_conn, ent_list)
    conf = format_basic_conf(ldap_conn, SCHEMA_RFC2307_BIS)
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)
    return None


def test_innetgr_with_threads(add_thread_test_netgroup):

    subprocess.check_call(["sss_netgroup_thread_test"])



""" IPA AD Trust Sanity tests

:requirement: IDM-SSSD-REQ: Testing SSSD in IPA Provider
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""

import random
import re
import time
from pexpect import pxssh
import pytest
from sssd.testlib.common.utils import sssdTools
from sssd.testlib.common.utils import ADOperations
from sssd.testlib.common.exceptions import SSHLoginException
from sssd.testlib.common.expect import pexpect_ssh


@pytest.mark.usefixtures('setup_ipa_client')
@pytest.mark.tier2
@pytest.mark.trust
class TestADTrust(object):
    """ IPA AD Trust tests """
    @staticmethod
    def test_basic_sssctl_list(multihost):
        """
        :title: Verify sssctl lists trusted domain
        :id: 8da8919d-524c-4498-8dc8-608eb5e139b0
        """
        domain_list = 'sssctl domain-list'
        ad_domain_name = multihost.ad[0].domainname
        cmd = multihost.master[0].run_command(domain_list, raiseonerr=False)
        mylist = cmd.stdout_text.split()
        assert ad_domain_name in mylist

    @staticmethod
    def test_pam_sss_gss_handle_large_krb_ticket(
            multihost, create_aduser_group):
        """
        :title: Verify pam_sss_gss.so can handle large kerberos ticket
                for sudo
        :id: 456ea53b-6702-4b8e-beb1-eee841b85fed
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1948657
        :steps:
         1. Add sudo rule in IPA-server for AD-users
         2. Modify /etc/krb5.conf.d/kcm_default_ccache to specify location
            of storing a TGT
         3. Enable pam_sss_gss.so for auth in /etc/pam.d/{sudo,sudo-i} files
         4. Add a sudo rule for AD-user
         5. Log in on ipa-client as AD-user
         6. Run kinit and fetch tgt
         7. Run sudo command
         8. Remove sudo cache
         9. Run sudo command again
        :expectedresults:
         1. Should succeed
         2. Should succeed
         3. Should succeed
         4. Should succeed
         5. Should succeed
         6. Should succeed
         7. Should not ask password, and should succeed
         8. Should succeed
         9. Should not ask password, and should succeed

        """
        # pylint: disable=too-many-locals, too-many-statements
        (aduser, _) = create_aduser_group
        ad_dmn_name = multihost.ad[0].domainname
        fq_aduser = f'{aduser}@{ad_dmn_name}'
        client = sssdTools(multihost.client[0], multihost.ad[0])
        ipaserver = sssdTools(multihost.master[0])
        cmd = 'dnf install -y sssd sssd-kcm'
        multihost.client[0].run_command(cmd, raiseonerr=False)
        domain_name = ipaserver.get_domain_section_name()
        domain_section = 'domain/{}'.format(domain_name)
        params = {'pam_gssapi_services': 'sudo, sudo-i'}
        client.sssd_conf(domain_section, params)
        krbkcm = '/etc/krb5.conf.d/kcm_default_ccache'
        bk_krbkcm = '/tmp/kcm_default_ccache'
        multihost.client[0].run_command(f'cp {krbkcm} {bk_krbkcm}')
        cmd = "echo -e  '[libdefaults]\n' \
              '    default_ccache_name  = FILE:/tmp/krb5cc_%{uid}:'"
        multihost.client[0].run_command(cmd, raiseonerr=False)
        multihost.client[0].service_sssd('restart')
        pam_sss_gss = "auth       sufficient   pam_sss_gss.so debug"
        for pam_file in "/etc/pam.d/sudo-i", "/etc/pam.d/sudo":
            cmd = f'sed -i "1 i {pam_sss_gss}" {pam_file}'
            multihost.client[0].run_command(cmd, raiseonerr=False)
        cmd = f'echo "{fq_aduser} ALL=(ALL) ALL" >> /etc/sudoers'
        multihost.client[0].run_command(cmd, raiseonerr=False)
        log = re.compile('.*System.*error.*Broken.*pipe.*')

        p_ssh = pxssh.pxssh(
            options={"StrictHostKeyChecking": "no",
                     "UserKnownHostsFile": "/dev/null"}
        )
        p_ssh.force_password = True
        try:
            p_ssh.login(multihost.client[0].ip, fq_aduser, 'Secret123')
            p_ssh.sendline(f'kinit {fq_aduser}')
            p_ssh.expect('Password for .*:', timeout=10)
            p_ssh.sendline('Secret123')
            p_ssh.prompt(timeout=5)
            p_ssh.sendline('sudo -l')
            p_ssh.prompt(timeout=5)
            sudo_out_1 = str(p_ssh.before)
            p_ssh.sendline('sudo id; echo "retcode:$?"')
            p_ssh.prompt(timeout=5)
            sudo_out_2 = str(p_ssh.before)
            p_ssh.sendline('sudo -l; echo "retcode:$?"')
            p_ssh.prompt(timeout=5)
            sudo_out_3 = str(p_ssh.before)
            p_ssh.logout()
        except pxssh.ExceptionPxssh:
            pytest.fail("Failed to login via ssh.")

        result = True
        for line in sudo_out_1.splitlines():
            res = log.search(line)
            result = result and res is None
        client.sssd_conf(domain_section, params, action='delete')
        for pam_file in "/etc/pam.d/sudo-i", "/etc/pam.d/sudo":
            cmd = f'sed -i "1d" {pam_file}'
            multihost.client[0].run_command(cmd, raiseonerr=False)
        cmd = 'sed -i "$ d" /etc/sudoers'
        multihost.client[0].run_command(cmd, raiseonerr=False)
        cmd = f'mv {bk_krbkcm} {krbkcm}'
        multihost.client[0].run_command(cmd, raiseonerr=False)
        assert result is True, "Error occurred with large ticket!"
        assert "retcode:0" in sudo_out_2, "sudo id failed"
        assert "retcode:0" in sudo_out_3, "sudo -l failed"

    @staticmethod
    def test_adusrname_beginning_with_at_rate_sign(multihost):
        """
        :title: AD_username beginning with @ sign
        :id: 1f7a3199-86a7-49f0-bfb2-69e81dbce244
        :customerscenario: True
        :bugzilla:
         https://bugzilla.redhat.com/show_bug.cgi?id=2180981
        :description:
         In the IPA-AD trust, AD-username starting  with '@' sign should be
         handled correctly.
        :setup:
          1. Create an AD-user
          2. Create an AD-group and add previously created AD-user as it's member
        :steps:
          1. From ipaclient, run id @<AD-user>
          2. From ipaclient, run getent passwd @<AD-user>
        :expectedresults:
          1. No user information should be returned
          2. No user information should be returned
        """
        client = sssdTools(multihost.client[0], multihost.ad[0])
        ad = ADOperations(multihost.ad[0])
        ad_dmn = multihost.ad[0].domainname
        ad_user = 'aduser7'
        ad_group = 'adgrp7'
        ad.create_ad_unix_user_group(ad_user, ad_group)
        client.clear_sssd_cache()
        cmd = multihost.client[0].run_command(
            f'id @{ad_user}@{ad_dmn}', raiseonerr=False)
        cmd1 = multihost.client[0].run_command(
            f'getent passwd @{ad_group}@{ad_dmn}', raiseonerr=False)
        ad.delete_ad_user_group(ad_user)
        ad.delete_ad_user_group(ad_group)
        assert cmd.returncode != 0, 'User information returned'
        assert ad_user not in cmd.stdout_text,\
            f"{ad_user} information is fetched"
        assert cmd1.returncode != 0, 'User information returned'
        assert ad_group not in cmd1.stdout_text,\
            f"{ad_user} is not available in {ad_user} information"

    @staticmethod
    def test_adgrpwith_at_ratesign(multihost):
        """
        :title: user membership of AD group with @ sign
        :id: ee9ca809-6ea7-48f2-a0fe-d9eccadf5d81
        :customerscenario: True
        :bugzilla:
         https://bugzilla.redhat.com/show_bug.cgi?id=2061795
        :description:
         In the IPA-AD trust, IPA-user having group membership of an
         AD-group, containing '@' sign in it's name, should be resolvable
         with default re_expression on a ipa-client system.
        :steps:
          1. Create an AD-group having '@' sign in it's name
          2. Create an AD-user and add it to above created group
          3. From ipaclient, ad-group with '@' sign is correctly fetched
          4. From ipaclient, confirm ad-user is showing correct group
             membership
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Should succeed
        """
        client = sssdTools(multihost.client[0], multihost.ad[0])
        ad = ADOperations(multihost.ad[0])
        ad_dmn = multihost.ad[0].domainname
        ad_user = 'aduser7'
        ad_group = 'adgrp@7'
        ad.create_ad_unix_user_group(ad_user, ad_group)
        client.clear_sssd_cache()
        cmd = multihost.client[0].run_command(
            f'getent group {ad_group}@{ad_dmn}', raiseonerr=False)
        cmd1 = multihost.client[0].run_command(
            f'id {ad_user}@{ad_dmn}', raiseonerr=False)
        ad.delete_ad_user_group(ad_user)
        ad.delete_ad_user_group(ad_group)
        assert ad_group in cmd.stdout_text,\
            f"{ad_group} information is fetched correctly"
        assert ad_group in cmd1.stdout_text,\
            f"{ad_group} is available in {ad_user} information"

    @staticmethod
    def test_ipaserver_sss_cache_user(multihost):
        """
        :title: Verify AD user is cached on IPA server
         when ipa client queries AD User
        :id: 4a48ee7a-62d1-4eea-9f33-7df3fccc908e
        """
        ipaserver = sssdTools(multihost.master[0])
        domain_name = ipaserver.get_domain_section_name()
        cache_path = '/var/lib/sss/db/cache_%s.ldb' % domain_name
        ad_domain_name = multihost.ad[0].domainname
        user_name = 'Administrator@%s' % ad_domain_name
        id_cmd = 'id %s' % user_name
        multihost.master[0].run_command(id_cmd, raiseonerr=False)
        multihost.client[0].run_command(id_cmd, raiseonerr=False)
        d_n = f'name=Administrator@{ad_domain_name},cn=users,' \
              f'cn={ad_domain_name},cn=sysdb'
        ldb_cmd = 'ldbsearch -H %s -b "%s"' % (cache_path, d_n)
        multihost.master[0].run_command(ldb_cmd, raiseonerr=False)

    @staticmethod
    def test_enforce_gid(multihost):
        """
        :title: Verify whether the new gid is enforceable when
         gid of AD Group Domain Users is overridden
        :id: 3581c7c0-d598-4e34-bb9b-9d791b93ec65
        :bugzilla:
         https://bugzilla.redhat.com/show_bug.cgi?id=1817219
        """
        create_view = 'ipa idview-add  foo_bar'
        multihost.master[0].run_command(create_view)
        ad_domain_name = multihost.ad[0].domainname
        ad_grp = 'Domain Users@%s' % ad_domain_name
        cmd = 'ipa idoverridegroup-add foo_bar "%s" --gid=40000000' % ad_grp
        multihost.master[0].run_command(cmd, raiseonerr=False)
        # apply the view on client
        client_hostname = multihost.client[0].sys_hostname
        apply_view = "ipa idview-apply foo_bar --hosts=%s" % client_hostname
        multihost.master[0].run_command(apply_view)
        client = sssdTools(multihost.client[0])
        client.clear_sssd_cache()
        time.sleep(5)
        user_name = 'Administrator@%s' % ad_domain_name
        id_cmd = 'id %s' % user_name
        cmd = multihost.client[0].run_command(id_cmd, raiseonerr=False)
        group = "40000000(domain users@%s)" % ad_domain_name
        delete_id_view = 'ipa idview-del foo_bar'
        multihost.master[0].run_command(delete_id_view)
        client.clear_sssd_cache()
        assert group in cmd.stdout_text

    @staticmethod
    def test_honour_idoverride(multihost, create_aduser_group):
        """
        :title: Verify sssd honours the customized ID View
        :id: 0c0dcfbb-6099-4c61-81c9-3bd3a003ff58
        :bugzilla:
         https://bugzilla.redhat.com/show_bug.cgi?id=1826720
        """
        (aduser, _) = create_aduser_group
        domain = multihost.ad[0].domainname
        ipa_client = sssdTools(multihost.client[0])
        ipa_client.clear_sssd_cache()
        ad_user_fqdn = '%s@%s' % (aduser, domain)
        id_cmd = 'id -g %s' % ad_user_fqdn
        cmd = multihost.master[0].run_command(id_cmd, raiseonerr=False)
        current_gid = cmd.stdout_text.strip()
        create_view = 'ipa idview-add madrid_trust_view'
        multihost.master[0].run_command(create_view)
        cmd = 'ipa idoverrideuser-add madrid_trust_view '\
              '%s --uid=50001 --gidnumber=50000 '\
              '--home=/home/%s' % (ad_user_fqdn, aduser)
        multihost.master[0].run_command(cmd, raiseonerr=False)
        # apply the view on client
        apply_view = "ipa idview-apply madrid_trust_view "\
                     "--hosts=%s" % multihost.client[0].sys_hostname
        multihost.master[0].run_command(apply_view)
        ipa_client.clear_sssd_cache()
        time.sleep(5)
        id_cmd = 'id %s' % ad_user_fqdn
        count = 0
        for _ in range(50):
            cmd = multihost.client[0].run_command(id_cmd, raiseonerr=False)
            gid = cmd.stdout_text.strip()
            if gid == current_gid:
                count += 1
        delete_id_view = 'ipa idview-del madrid_trust_view'
        multihost.master[0].run_command(delete_id_view)
        ipa_client.clear_sssd_cache()
        assert count == 0

    @staticmethod
    def test_ipa_missing_secondary_ipa_posix_groups(multihost,
                                                    create_aduser_group):
        """
        :title: IPA missing secondary IPA Posix groups in latest sssd
        :id: bbb82516-4127-4053-9b06-9104ac889819
        :setup:
         1. Configure trust between IPA server and AD.
         2. Configure client machine with SSSD integrated to IPA.
         3. domain-resolution-order set so the AD domains are checked first
         4. Create external group that is member of a posix group
         5. Create user that is a member of the external group
         6. Make sure that external group is member of posix group.
        :steps:
         0. Clean sssd cache
         1. Run getent group for posix group and using id check that user
            is member of posix group.
        :expectedresults:
         0. Cache is cleared.
         1. The posix group gid is present in id output.
        :teardown:
         Remove the created user, groups and revert resolution order.
        :customerscenario: True
        :bugzilla:
         https://bugzilla.redhat.com/show_bug.cgi?id=1945552
         https://bugzilla.redhat.com/show_bug.cgi?id=1937919
         https://bugzilla.redhat.com/show_bug.cgi?id=1945654
        """
        # pylint: disable=too-many-locals
        ad_domain = multihost.ad[0].domainname
        ipaserver = sssdTools(multihost.master[0])
        ipa_domain = ipaserver.get_domain_section_name()
        (username, _) = create_aduser_group
        posix_group = "posix_group_01"
        ext_group = "ext_group_01"
        # SETUP
        # Set the domain resolution order to AD first
        resorder_cmd = f'ipa config-mod --domain-resolution-order=' \
                       f'{ad_domain}:{ipa_domain}'
        multihost.master[0].run_command(resorder_cmd, raiseonerr=False)

        # Create posix group
        pgroup_cmd = f'ipa group-add {posix_group}'
        multihost.master[0].run_command(pgroup_cmd, raiseonerr=False)

        # Create and external group
        ext_group_cmd = f'ipa group-add --external {ext_group}'
        multihost.master[0].run_command(ext_group_cmd, raiseonerr=False)

        # Set membership of external group in posix group
        member_cmd = f'ipa -n group-add-member {posix_group} --groups=' \
                     f'{ext_group}'
        multihost.master[0].run_command(member_cmd, raiseonerr=False)

        # Set AD user membership in external group
        usr_mbr_cmd = f"ipa -n group-add-member {ext_group} --external" \
                      f" '{username}@{ad_domain}'"
        multihost.master[0].run_command(usr_mbr_cmd, raiseonerr=False)

        # TEST
        # Get posix group id
        grp_show_cmd = f"ipa group-show {posix_group}"
        cmd = multihost.master[0].run_command(grp_show_cmd, raiseonerr=False)
        gid_regex = re.compile(r"GID: (\d+)")
        posix_group_id = gid_regex.search(cmd.stdout_text).group(1)

        # Check that external group is member of posix group
        grp_show_cmd = f"ipa group-show {ext_group}"
        cmd = multihost.master[0].run_command(grp_show_cmd, raiseonerr=False)
        assert posix_group in cmd.stdout_text, \
            "The external group is not a member of posix group!"

        # A bit of wait so the user is propagated
        time.sleep(60)

        # The reproduction rate is not 100%, I had reliably 2+
        # fails in 5 rounds.
        for _ in range(5):
            # Clean caches on SSSD so we don't have to wait for cache timeouts
            # The reproduction works better on sssd on ipa master
            sssd_client = sssdTools(multihost.master[0])
            sssd_client.clear_sssd_cache()

            # Search the posix group using getent to trigger the condition with
            # negative cache
            getent_cmd = f"getent group {posix_group_id}"
            multihost.master[0].run_command(getent_cmd, raiseonerr=False)

            # Check that posix group is listed in id
            id_cmd = f"id {username}@{ad_domain}"
            cmd = multihost.master[0].run_command(id_cmd, raiseonerr=False)
            # Check if id worked
            assert cmd.returncode == 0,\
                'Could not find the user, something wrong with setup!'
            # Check if the posix group was found for the user.
            assert posix_group_id in cmd.stdout_text,\
                "The user is not a member of posix group!"

        # TEARDOWN
        # Remove user from external group
        usr_mbr_del_cmd = f"ipa -n group-remove-member {ext_group} " \
                          f"--external '{username}@{ad_domain}'"
        multihost.master[0].run_command(usr_mbr_del_cmd, raiseonerr=False)

        # Remove group membership
        grp_del_mbr_cmd = f'ipa -n group-remove-member {posix_group}' \
                          f' --groups={ext_group}'
        multihost.master[0].run_command(grp_del_mbr_cmd, raiseonerr=False)

        # Remove external group
        ext_grp_del_cmd = f'ipa group-del {ext_group}'
        multihost.master[0].run_command(ext_grp_del_cmd, raiseonerr=False)

        # Remove posix group
        px_grp_del_cmd = f'ipa group-del {posix_group}'
        multihost.master[0].run_command(px_grp_del_cmd, raiseonerr=False)

        # Reset the domain resolution order
        rev_resorder_cmd = f'ipa config-mod --domain-resolution-order=' \
                           f'{ipa_domain}:{ad_domain}'
        multihost.master[0].run_command(rev_resorder_cmd, raiseonerr=False)

    @staticmethod
    def test_nss_get_by_name_with_private_group(multihost):
        """
        :title:
         SSSD fails nss_getby_name for IPA user with SID if the user has
         a private group
        :id: 45dce6b9-0d47-4b9f-9532-4da8178e5334
        :setup:
         1. Configure trust between IPA server and AD.
         2. Configure client machine with SSSD integrated to IPA.
         3. Create an user with a private group
        :steps:
         1. Call function getsidbyname from pysss_nss_idmap for admin.
         2. Call function getsidbyname from pysss_nss_idmap for then user.
        :expectedresults:
         1. The admin SID is returned.
         2. The user SID is returned.
        :teardown:
         Remove the created user.
        :bugzilla:
         https://bugzilla.redhat.com/show_bug.cgi?id=1837090
        """
        # Create an user with a private group
        username = 'some-user'
        multihost.master[0].run_command(
            f'ipa user-add {username} --first=Some --last=User',
            raiseonerr=False
        )

        # Confirm that the user exists
        cmd = multihost.master[0].run_command(
            f'id  {username}',
            raiseonerr=False
        )
        # First check for admin user to make sure that the setup is correct
        check_admin_cmd = '''python3 -c "import pysss_nss_idmap; import '''\
            '''sys; result=pysss_nss_idmap.getsidbyname('admin');'''\
            '''print(result); result or sys.exit(2)"'''
        cmd_adm = multihost.master[0].run_command(check_admin_cmd,
                                                  raiseonerr=False)

        # Now check for the user with the private group
        check_user_cmd = '''python3 -c "import pysss_nss_idmap; import sys;'''\
            '''result=pysss_nss_idmap.getsidbyname('%s');print(result); '''\
            '''result or sys.exit(2)"''' % username
        cmd_usr = multihost.master[0].run_command(check_user_cmd,
                                                  raiseonerr=False)

        # Remove the user afterwards
        user_del_cmd = f'ipa user-del {username}'
        multihost.master[0].run_command(user_del_cmd, raiseonerr=False)

        # Evaluate results after cleanup is done
        assert cmd.returncode == 0, 'Could not find the user!'
        assert cmd_adm.returncode == 0, 'Something wrong with setup!'
        assert cmd_usr.returncode == 0, \
            f"pysss_nss_idmap.getsidbyname for {username} failed"

    @staticmethod
    def test_idview_override_group_fails(multihost, create_aduser_group):
        """
        :title: IPA clients fail to resolve override group names in custom view
        :id: 7a0dc871-fdad-4c07-9d07-a092baa83178
        :customerscenario: true
        :bugzilla:
          https://bugzilla.redhat.com/show_bug.cgi?id=2004406
          https://bugzilla.redhat.com/show_bug.cgi?id=2031729
        :description: Overriding both user and group names and ids in
          an idview for user and group from AD results in error in sssd
          when running id command.
        :setup:
          1. Create user and group (group1) on AD.
          2. Make AD user member of group1.
          3. Create additional group (group2) on AD.
        :steps:
          1. ID views to override AD groupname and gid of group1.
          2. ID views to override AD groupname and gid of group2.
          3. ID view to override AD username, uid and gid (to gid of group2).
          4. Run an "id" command for the override user.
        :expectedresults:
          1. View with an override is created.
          2. View with an override is created.
          3. User override is added to the view.
          4. Id command succeeds, group override is visible, all groups are
             properly resolved.
        """
        (aduser, adgroup) = create_aduser_group
        run_id_int = random.randint(9999, 999999)
        adgroup2 = f"group2_{run_id_int}"
        ado = ADOperations(multihost.ad[0])
        ado.create_ad_unix_group(adgroup2)
        domain = multihost.ad[0].domainname

        ipa_client = sssdTools(multihost.client[0])
        ipa_client.clear_sssd_cache()

        view = f'prygl_trust_view_{run_id_int}'
        create_view = f'ipa idview-add {view}'
        multihost.master[0].run_command(create_view, raiseonerr=False)

        create_grp_override = f'ipa idoverridegroup-add "{view}" ' \
            f'{adgroup}@{domain} --group-name ' \
            f'"borci{run_id_int}" --gid={run_id_int+1}'
        multihost.master[0].run_command(create_grp_override, raiseonerr=False)

        create_grp2_override = f'ipa idoverridegroup-add "{view}" ' \
            f'{adgroup2}@{domain} --group-name ' \
            f'"magori{run_id_int}" --gid={run_id_int+2}'
        multihost.master[0].run_command(create_grp2_override, raiseonerr=False)

        create_user_override = f'ipa idoverrideuser-add "{view}" ' \
            f'{aduser}@{domain} --login ferko{run_id_int} ' \
            f'--uid=50001 --gidnumber={run_id_int+2}'
        multihost.master[0].run_command(create_user_override, raiseonerr=False)

        # Apply the view on client
        multihost.master[0].run_command(
            f"ipa idview-apply '{view}' --hosts="
            f"{multihost.client[0].sys_hostname}", raiseonerr=False)

        ipa_client.clear_sssd_cache()
        time.sleep(5)
        cmd = multihost.client[0].run_command(
            f'id ferko{run_id_int}@{domain}', raiseonerr=False)

        # TEARDOWN
        ado.delete_ad_user_group(adgroup2)
        multihost.master[0].run_command(
            f'ipa idview-del {view}', raiseonerr=False)

        # Test result Evaluation
        assert cmd.returncode == 0, f"User {aduser} was not found."
        assert f"borci{run_id_int}@{domain}" in cmd.stdout_text,\
            f"Group 1 {adgroup} name was not overridden/resolved."
        assert f"magori{run_id_int}@{domain}" in cmd.stdout_text,\
            f"Group 2 {adgroup2} name was not overridden/resolved."
        assert f"{run_id_int+1}" in cmd.stdout_text,\
            "Group 1 id was not overridden."
        assert f"{run_id_int+2}" in cmd.stdout_text,\
            "Group 2 id was not overridden."
        assert f"domain users@{domain}" in cmd.stdout_text, \
            "Group domain users is missing."

    @staticmethod
    def test_ad_user_ssh_ipa_client(multihost):
        """
        :title: Cannot SSH with AD user to ipa-client
        :description: 'krb5_validate' and 'pac_check' settings conflicted
          before the fix. By default 'krb5_validate = true' with id_providers
          ipa and ad_provider. Setting it to false will enable login even if
          the ticket validation does not work
        :id: d4af4084-0ee2-4339-be66-ef117439f32d
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=2128544
                   https://bugzilla.redhat.com/show_bug.cgi?id=2128902
        :steps:
          1. Add an invalid entry keytab after taking a backup
          2. ssh as AD user on the ipa client machine
          3. Set 'krb5_validate = false', debug_level 9 in sssd.conf,
             and restart sssd service
          4. ssh as AD user on the ipa client machine
          5. Check for log message on PAC check being skipped with
             krb_validate set to false
        :expectedresults:
          1. Invalid entry is added at the end of the keytab
          2. AD user ssh login fails
          3. sssd restart successfully with 'krb5_validate = false'
          4. AD user ssh login is successful with the setting
          5. Log message about skipping pac_check is found
        """
        ad_domain_name = multihost.ad[0].domainname
        client_hostip = multihost.client[0].ip
        multihost.client[0].run_command('cp -f /etc/krb5.keytab '
                                        '/etc/krb5.keytab.bak', raiseonerr=False)
        ktutil_cmd = '(echo "rkt /etc/krb5.keytab"; sleep 1; echo "addent ' \
                     '-password -p invalid@invaliddom -k 3 -e ' \
                     'aes128-cts-hmac-sha1-96"; sleep 1; echo "pass00189";' \
                     'sleep 1; echo "list"; sleep 1; echo ' \
                     '"wkt /tmp/invalid.keytab"; sleep 1; echo "quit";) | ktutil'

        multihost.client[0].run_command(ktutil_cmd, raiseonerr=False)
        multihost.client[0].run_command('mv /tmp/invalid.keytab '
                                        '/etc/krb5.keytab', raiseonerr=False)
        multihost.client[0].run_command('restorecon /etc/krb5.keytab', raiseonerr=False)
        multihost.client[0].run_command('klist -ekt /etc/krb5.keytab', raiseonerr=False)

        client = pexpect_ssh(client_hostip, f'user1@{ad_domain_name}',
                             'Secret123', debug=True)
        with pytest.raises(Exception):
            client.login(login_timeout=10, sync_multiplier=1,
                         auto_prompt_reset=False)

        tools = sssdTools(multihost.client[0])
        domain_params = {'debug_level': '9', 'krb5_validate': 'false'}
        ipaserver = sssdTools(multihost.master[0])
        domain_name = ipaserver.get_domain_section_name()
        tools.sssd_conf(f'domain/{domain_name}', domain_params)
        tools.clear_sssd_cache()

        client = pexpect_ssh(client_hostip, f'user1@{ad_domain_name}',
                             'Secret123', debug=True)
        try:
            client.login(login_timeout=30, sync_multiplier=2,
                         auto_prompt_reset=False)
        except SSHLoginException:
            pytest.fail("user failed to login")
        else:
            client.logout()

        multihost.client[0].run_command('cp -f /etc/krb5.keytab.bak '
                                        '/etc/krb5.keytab', raiseonerr=False)
        multihost.client[0].run_command('restorecon /etc/krb5.keytab', raiseonerr=False)

        log_str = 'PAC check is requested but krb5_validate is '\
                  'set to false. PAC checks will be skipped'

        multihost.client[0].run_command(f'grep -i "{log_str}" '
                                        '/var/log/sssd/krb5_child.log')
        multihost.client[0].run_command(f'grep -i "{log_str}" '
                                        f'/var/log/sssd/sssd_{domain_name}.log')

    @staticmethod
    def test_skip_members_in_view_search(multihost):
        """
        :title: Skip group members that point to an entry in the views base search
        :id: 40adfa6a-b951-4aab-aa5b-57e149526e8d
        :customerscenario: true
        :bugzilla:
          https://bugzilla.redhat.com/show_bug.cgi?id=2151403
        :description: When resolving the nested groups, any entry in
         cn=views,cn=accounts,$BASEDN (or whatever the user configured
         using ipa_views_search_base) must be ignored.
        :setup:
         1. Configure trust between IPA server and AD.
         2. Configure client machine with SSSD integrated to IPA.
        :steps:
          1. Add a user1 ID override in 'default trust view'.
          2. Create ext_group as a external group and add user in admins and external group.
          3. Make ext_group a member of the admins group.
          4. Run an "id" command for the override user.
        :expectedresults:
          1. Successfully added user1 in default trust view.
          2. Successfully created and added user in admins and external group.
          3. Made external_group a member of the admins group.
          4. Id command succeeds, admin group is visible, groups are properly resolved.
        """
        kinit_admin = 'kinit admin'
        multihost.master[0].run_command(kinit_admin, stdin_text='Secret123',
                                        raiseonerr=False)

        domain = multihost.ad[0].domainname

        ipa_client = sssdTools(multihost.client[0])
        ipa_master = sssdTools(multihost.master[0])
        ipa_client.clear_sssd_cache()

        # Here user1 is already added in AD and also member of group1, group2 and group3
        # Add a new User ID override in default trust view.
        create_newuser_override = f"ipa idoverrideuser-add 'default trust view' user1@{domain}"
        multihost.master[0].run_command(create_newuser_override, raiseonerr=False)

        # Add user to ‘admins’ group
        add_user_override = f"ipa group-add-member admins --idoverrideusers user1@{domain}"
        multihost.master[0].run_command(add_user_override, raiseonerr=False)

        # Add a external group named as a ext_group
        create_ext_grp = f'ipa group-add --external ext_group'
        multihost.master[0].run_command(create_ext_grp, raiseonerr=False)

        # Add a user in external group named as a ext_group
        add_grp_members = f'ipa -n group-add-member ext_group --external=user1@{domain}'
        multihost.master[0].run_command(add_grp_members, raiseonerr=False)

        # Make ext_group a member of the admins group, as a result the members of ext_group will be
        # members of admin too.
        add_ext_grp_to_admin = f'ipa group-add-member admins --groups=ext_group'
        multihost.master[0].run_command(add_ext_grp_to_admin, raiseonerr=False)

        ipa_master.clear_sssd_cache()
        ipa_client.clear_sssd_cache()
        time.sleep(5)

        lookup_id_master = multihost.master[0].run_command(f'id user1@{domain}', raiseonerr=False)
        lookup_id_client = multihost.client[0].run_command(f'id user1@{domain}', raiseonerr=False)

        # Teardown the setup and delete the external group
        cmd_to_delete = ["ipa group-remove-member admins --groups=ext_group",
                         f"ipa -n group-remove-member ext_group --external=user1@{domain}",
                         "ipa group-del ext_group",
                         f"ipa group-remove-member admins --idoverrideusers user1@{domain}",
                         f"ipa idoverrideuser-del 'default trust view' user1@{domain}"]
        for cmd in cmd_to_delete:
            multihost.master[0].run_command(cmd, raiseonerr=False)

        # Check the lookup from ips-server first.
        assert lookup_id_master.returncode == 0, f"Lookup is failed for user1@{domain} on ipa-server"
        assert f"admins" in lookup_id_master.stdout_text, "admins name was not resolved on ipa-server."

        # Check the lookup from sssd client.
        assert lookup_id_client.returncode == 0, f"Lookup is failed for user1@{domain} on ipa-client"
        assert f"group1@{domain}" in lookup_id_client.stdout_text, f"group1 name was not resolved" \
                                                                   f" on ipa-client."
        assert f"admins" in lookup_id_client.stdout_text, "admins name was not resolved on ipa-client."

    @staticmethod
    def test_search_filter_for_override_usrgrp(multihost, create_aduser_group):
        """
        :title: BE_REQ_USER_AND_GROUP LDAP search filter can inadvertently catch multiple overrides
        :id: 932f1177-2375-46a6-8920-9f06d8874881
        :customerscenario: true
        :bugzilla:
          https://bugzilla.redhat.com/show_bug.cgi?id=2096183
        :description: The new filter looks for a specific user override or a specific
        group override:(|(&(objectClass=ipaUserOverride)(uidNumber=XXXX))
        (&(objectClass=ipaGroupOverride)(gidNumber=XXXX))).
        :setup:
          1. Create user and group (group1) on AD.
          2. Make AD user member of group1.
        :steps:
          1. Add a group override to the 'default trust view' with gid number and group name.
          2. Add a user override to the 'default trust view' view with same gid number
             which was created in step 1.
          3. Lookup sid-by-id with pysss_nss_idmap using python.
          4. Lookup for the group and again run step 3.
          5. Check the log message from domain log
        :expectedresults:
          1. Group added in 'default trust view' with gid number and group name.
          2. User added in 'default trust view' with same gid number which was created in step 1.
          3. Successfully lookup sid-by-id with pyss_nss_idmap using python.
          4. Group lookup command succeeds, successfully ran step 3 again.
          5. Got expected log messages from domain log.
        """
        (aduser, adgroup) = create_aduser_group
        run_id_int = random.randint(9999, 999999)
        domain = multihost.ad[0].domainname

        ipa_client = sssdTools(multihost.client[0])
        ipa_master = sssdTools(multihost.master[0])

        add_grp_override = f'ipa idoverridegroup-add "default trust view" {adgroup}@{domain} ' \
                           f'--group-name "borci{run_id_int}" --gid={run_id_int}'
        multihost.master[0].run_command(add_grp_override, raiseonerr=False)

        add_user_override = f'ipa idoverrideuser-add "default trust view" {aduser}@{domain} ' \
                            f'--login ferko{run_id_int} --uid=50001 --gidnumber={run_id_int}'
        multihost.master[0].run_command(add_user_override, raiseonerr=False)

        domain_params = {'debug_level': '9'}
        domain_name = ipa_client.get_domain_section_name()
        ipa_client.sssd_conf(f'domain/{domain_name}', domain_params)

        ipa_master.clear_sssd_cache()
        ipa_client.clear_sssd_cache()
        time.sleep(5)

        # Lookup sid-by-id with pysss_nss_idmap using python
        python_cmd = f'''python3 -c "import pysss_nss_idmap; print (pysss_nss_idmap.getsidbyid('{run_id_int}'))"'''
        multihost.client[0].run_command(python_cmd, raiseonerr=False)

        group_lookup = f'getent group {adgroup}@{domain}'
        check_gr_lookup = multihost.client[0].run_command(group_lookup, raiseonerr=False)

        multihost.client[0].run_command(python_cmd, raiseonerr=False)

        # Updated log message from sssd domain log
        log_message = f'Found override for object with filter'

        # Download sssd log
        log_file = multihost.client[0].get_file_contents(f"/var/log/sssd/sssd_{domain_name}.log").decode('utf-8')

        # Teardown the setup
        cmd_to_delete = [f"ipa idoverridegroup-del 'default trust view' {adgroup}@{domain}",
                         f"ipa idoverrideuser-del 'default trust view' {aduser}@{domain}"]
        for cmd in cmd_to_delete:
            multihost.master[0].run_command(cmd, raiseonerr=False)

        # Test result Evaluations
        assert check_gr_lookup.returncode == 0, f"group {adgroup} was not found."
        assert f"borci{run_id_int}@{domain}" in check_gr_lookup.stdout_text, "Group name was not resolved."
        assert f"ferko{run_id_int}@{domain}" in check_gr_lookup.stdout_text, "Group name was not resolved."
        assert log_message in log_file

    @staticmethod
    def test_ldap_user_extra_attrs(multihost, create_aduser_group):
        """
        :title: When adding attributes ldap_user_extra_attrs with mail value in sssd.conf
        the cross-forest query stop working
        :id: abc60b52-224d-4ac3-bbae-195cb0c563a0
        :customerscenario: true
        :bugzilla:
          https://bugzilla.redhat.com/show_bug.cgi?id=2170720
        :description: When adding attributes ldap_user_extra_attrs in sssd.conf
          with the cross-forest, id command failed on client side.
        working.
        :setup:
          1. Create user and group on AD.
        :steps:
          1. Add ldap_user_extra_attrs in domain section with mail in client and master.
          2. Add user_attributes in nss section with mail in client and master.
          3. Clear sssd cache from client and master.
          4. Check id lookup of the user.
        :expectedresults:
          1. Successfully add parameter in domain section of client and master.
          2. Successfully add user_attributes in nss section of client and master.
          3. Cleared the cache of client and master.
          4. User lookup command succeeds.
        """
        client = sssdTools(multihost.client[0], multihost.ad[0])
        master = sssdTools(multihost.master[0])
        domain = multihost.ad[0].domainname

        (aduser, adgroup) = create_aduser_group
        section = client.get_domain_section_name()

        for role in [client, master]:
            domain_params = {'ldap_user_extra_attrs': f'mail, lastname:sn, firstname:givenname'}
            nss_params = {'user_attributes': '+mail, +firstname, +lastname'}
            role.sssd_conf(f'domain/{section}', domain_params)
            role.sssd_conf('nss', nss_params)
            role.clear_sssd_cache()

        # Test evaluation
        id_lookup = f'id {aduser}@{domain}'
        check_id = multihost.client[0].run_command(id_lookup, raiseonerr=False)

        assert check_id.returncode == 0, f'{aduser} id is not successful'
        assert f"{aduser}@{domain}" in check_id.stdout_text, "User name was not resolved."
        assert f"{adgroup}@{domain}" in check_id.stdout_text, "Group name was not resolved."

    @staticmethod
    def test_s2n_get_request(multihost):
        """
        :title: User lookup on IPA client fails with 's2n get_fqlist request failed'
        :id: fae73dd7-5a18-4aa4-a39f-a20a2f66b1c9
        :customerscenario: true
        :bugzilla:
            https://bugzilla.redhat.com/show_bug.cgi?id=2192708
            https://bugzilla.redhat.com/show_bug.cgi?id=2196838
            https://bugzilla.redhat.com/show_bug.cgi?id=2196816
            https://bugzilla.redhat.com/show_bug.cgi?id=2196839
        :description: When checking if the input group-name of an autogenerated
            user-private-group is the original name from AD or an overwritten one the
            comparison is currently done case sensitive.
            Since AD handles names case-insensitive and hence SSSD should do this as
            well to make sure that e.g. mixed-case names like Administrator can be
            match reliable.
        :setup:
          1. Add an Administrator user override to the 'default trust view' with sshpubkey.
        :steps:
          1. Check group lookup for Administrator user using mixed chars upper/lower cases.
        :expectedresults:
          1. Successfully group lookup the administrator user using mixed chars cases.
        """
        domain = multihost.ad[0].domainname

        ipa_client = sssdTools(multihost.client[0])
        ipa_master = sssdTools(multihost.master[0])

        ssh_key = f"ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQCiOfcg3WRL6z+6XWSAw4mT7q7aE7rj1KmhaM6U5" \
                  f"fmbN5QnEfCAMp8qbSbBLsjY0F501ZNpcAgeefDv3oNYL62sfac8OzWf5eXZlKJQYYILi8dv8i8HoJ" \
                  f"BT9+n81Y5w1UVbmRNX9n2lqqxdfhiL2iIsbzJ1KGmIw6JlmbeRtcgGRQzt0M+Ggftl6Kr97obEWo1" \
                  f"l9E5QWvkliecPXWJVTBUpM+Gr2CWhqLtNf5VALjYilX3jfC2355hIR8R8UsnkbWbjNksj7nruUQP9" \
                  f"goHcbJ6vbyzka3v/2aRC5eIa7b8NE8vwRbrtp5CV9QNbx/GiTY6T50CJE0lyEwmlLHKUmovt " \
                  f"Administrator@{domain}"

        # Add override user with ssh pub key
        add_user_override = f"ipa idoverrideuser-add 'Default Trust View' Administrator@{domain} " \
                            f"--sshpubkey='{ssh_key}'"
        multihost.master[0].run_command(add_user_override, raiseonerr=False)

        ipa_master.clear_sssd_cache()
        ipa_client.clear_sssd_cache()
        time.sleep(5)

        # check lookup of group
        group_lookup1 = f'getent group administrator@{domain}'
        check_gr_lookup1 = multihost.client[0].run_command(group_lookup1, raiseonerr=False)

        group_lookup2 = f'getent group adMiniStraTor@{domain}'
        check_gr_lookup2 = multihost.client[0].run_command(group_lookup2, raiseonerr=False)

        # Delete an Administrator User ID override
        cmd_to_delete = f"ipa idoverrideuser-del 'default trust view' administrator@{domain}"
        multihost.master[0].run_command(cmd_to_delete, raiseonerr=False)

        # Test result Evaluations
        assert check_gr_lookup1.returncode == 0 and check_gr_lookup2.returncode == 0, \
            f"group lookup was not resolved."
        assert f"administrator@{domain}" in check_gr_lookup1.stdout_text, "Group name was not resolved."
        assert f"administrator@{domain}" in check_gr_lookup2.stdout_text, "Group name was not resolved."



#
# Module for simulation of utility "getent networks -s sss" from coreutils
#
#   Authors:
#       Samuel Cabrero <scabrero@suse.com>
#
#   Copyright (C) 2020 SUSE LINUX GmbH, Nuernberg, Germany.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ctypes import (c_int, c_char_p, c_ulong, c_uint32, POINTER,
                    Structure, create_string_buffer)
from sssd_nss import NssReturnCode, SssdNssError, nss_sss_ctypes_loader
import socket
from ipaddress import IPv4Address
from struct import unpack

IP_NETWORK_BUFLEN = 1024


class Netent(Structure):
    _fields_ = [("n_name", c_char_p),
                ("n_aliases", POINTER(c_char_p)),
                ("n_addrtype", c_int),
                ("n_net", c_uint32)]


def getnetbyname_r(name, result_p, buffer_p, buflen):
    """
    ctypes wrapper for:
        enum nss_status _nss_sss_getnetbyname_r(const char *name,
                                                struct netent *result,
                                                char *buffer, size_t buflen,
                                                int *errnop, int *h_errnop)
    """
    func = nss_sss_ctypes_loader("_nss_sss_getnetbyname_r")
    func.restype = c_int
    func.argtypes = [c_char_p, POINTER(Netent),
                     c_char_p, c_ulong, POINTER(c_int), POINTER(c_int)]

    errno = POINTER(c_int)(c_int(0))
    h_errno = POINTER(c_int)(c_int(0))

    name = name.encode('utf-8')
    res = func(c_char_p(name), result_p, buffer_p, buflen, errno, h_errno)

    return (int(res), int(errno[0]), int(h_errno[0]), result_p)


def getnetbyaddr_r(addr, af, result_p, buffer_p, buflen):
    """
    ctypes wrapper for:
        enum nss_status _nss_sss_getnetbyaddr_r(uint32_t addr, int type,
                                                struct netent *result,
                                                char *buffer, size_t buflen,
                                                int *errnop, int *h_errnop)
    """
    func = nss_sss_ctypes_loader("_nss_sss_getnetbyaddr_r")
    func.restype = c_int
    func.argtypes = [c_uint32, c_int, POINTER(Netent),
                     c_char_p, c_ulong, POINTER(c_int), POINTER(c_int)]

    errno = POINTER(c_int)(c_int(0))
    h_errno = POINTER(c_int)(c_int(0))

    res = func(addr, af, result_p, buffer_p, buflen, errno, h_errno)

    return (int(res), int(errno[0]), int(h_errno[0]), result_p)


def set_netent_dict(res, result_p):
    if res != NssReturnCode.SUCCESS:
        return dict()

    netent_dict = dict()
    netent_dict['name'] = result_p[0].n_name.decode('utf-8')
    netent_dict['aliases'] = list()
    netent_dict['addrtype'] = result_p[0].n_addrtype
    netent_dict['address'] = result_p[0].n_net

    i = 0
    while result_p[0].n_aliases[i] is not None:
        alias = result_p[0].n_aliases[i].decode('utf-8')
        netent_dict['aliases'].append(alias)
        i = i + 1

    return netent_dict


def call_sssd_getnetbyname(name):
    """
    A Python wrapper to retrieve an IP network by name. Returns:
        (res, netent_dict)
    if res is NssReturnCode.SUCCESS, then netent_dict contains the keys
    corresponding to the C netent structure fields. Otherwise, the dictionary
    is empty and errno indicates the error code
    """
    result = Netent()
    result_p = POINTER(Netent)(result)
    buff = create_string_buffer(IP_NETWORK_BUFLEN)

    (res, errno, h_errno, result_p) = getnetbyname_r(name, result_p,
                                                     buff, IP_NETWORK_BUFLEN)
    if errno != 0:
        raise SssdNssError(errno, "getnetbyname_r")

    netent_dict = set_netent_dict(res, result_p)
    return (res, h_errno, netent_dict)


def call_sssd_getnetbyaddr(addrstr, af):
    """
    A Python wrapper to retrieve an IP network by address. Returns:
        (res, netent_dict)
    if res is NssReturnCode.SUCCESS, then netent_dict contains the keys
    corresponding to the C netent structure fields. Otherwise, the dictionary
    is empty and errno indicates the error code
    """
    result = Netent()
    result_p = POINTER(Netent)(result)
    buff = create_string_buffer(IP_NETWORK_BUFLEN)

    if isinstance(addrstr, bytes):
        addrstr = addrstr.decode('utf-8')
    addr = IPv4Address(addrstr)
    binaddr = unpack('<I', addr.packed)[0]
    binaddr = socket.ntohl(binaddr)

    (res, errno, h_errno, result_p) = getnetbyaddr_r(binaddr, af,
                                                     result_p, buff,
                                                     IP_NETWORK_BUFLEN)
    if errno != 0:
        raise SssdNssError(errno, "getnetbyaddr_r")

    netent_dict = set_netent_dict(res, result_p)
    return (res, h_errno, netent_dict)



#
#   Authors:
#       Pavel Brezina <pbrezina@redhat.com>
#
#   Copyright (C) 2017 Red Hat
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from collections import OrderedDict
from sbus_Introspection import SBus


class Invoker:
    """ Invoker is a piece of C code that takes care of executing a method,
        signal and property handlers and returning their output values.

        SBus code generator tries to reduce amount of generated code by
        reusing invokers whenever possible. Therefore we must ensure that
        invoker for each input and output signature type is generated only
        once.

        Each invoker is associated with its input and output SBus signatures
        extended by a reserved keyword "raw" that says that the invoker input
        or output parameters are handled by its handler and caller respectively
        and are passed through as D-Bus iterators.
    """
    def __init__(self, sbus_input, sbus_output):
        self.input = self.getSignature(sbus_input,
                                       self.IsCustomInputHandler(sbus_input))

        self.output = self.getSignature(
            sbus_output,
            self.IsCustomOutputHandler(sbus_output)
        )

    def getSignature(self, sbus_signature, is_custom_handler):
        if sbus_signature is None:
            return InvokerSignature("", {}, {})

        invoker_signature = sbus_signature.signature
        if is_custom_handler:
            invoker_signature = "raw"

        return InvokerSignature(invoker_signature,
                                sbus_signature.arguments,
                                sbus_signature.annotations)

    @staticmethod
    def GatherInvokers(interfaces):
        """
            Gather all required invokers for given interfaces.
        """
        dict = {}
        for iface in interfaces.values():
            for method in iface.methods.values():
                Invoker.Add(dict, method.input, method.output)

            for signal in iface.signals.values():
                Invoker.Add(dict, signal.input, signal.output)

            for property in iface.properties.values():
                if property.isReadable():
                    Invoker.Add(dict, None, property.output)
                if property.isWritable():
                    Invoker.Add(dict, property.input, None)

        return OrderedDict(sorted(dict.items()))

    @staticmethod
    def Add(dict, input, output):
        """
            Add a new invoker to dictionary if possible.
        """
        invoker = Invoker(input, output)
        key = "in:%s, out:%s" % (invoker.input.invokerSignature,
                                 invoker.output.invokerSignature)
        if key in dict:
            return

        dict[key] = invoker

    @staticmethod
    def IsCustomHandler(type, sbus_signature):
        if type == "input":
            return Invoker.IsCustomInputHandler(sbus_signature)
        elif type == "output":
            return Invoker.IsCustomOutputHandler(sbus_signature)
        else:
            raise ValueError("Invalid type: %s" % type)

    @staticmethod
    def IsCustomInputHandler(sbus_signature):
        names = ["codegen.CustomHandler",
                 "codegen.CustomInputHandler"]

        if sbus_signature is None:
            return False

        return SBus.Annotation.CheckIfTrue(names, sbus_signature.annotations)

    @staticmethod
    def IsCustomOutputHandler(sbus_signature):
        names = ["codegen.CustomHandler",
                 "codegen.CustomOutputHandler"]

        if sbus_signature is None:
            return False

        return SBus.Annotation.CheckIfTrue(names, sbus_signature.annotations)


class InvokerSignature:
    """ Contains information about Invoker signature and SBus arguments
        and annotations. Do not confuse with SBus.Signature.
    """
    def __init__(self, invoker_signature, sbus_arguments, sbus_annotations):
        self.invokerSignature = invoker_signature
        self.arguments = sbus_arguments
        self.annotations = sbus_annotations


class InvokerArgumentType:
    """ Argument reader/writer is a piece of C code that takes care of
        parsing D-Bus methods into C types.

        SBus code generator tries to reduce amount of generated code by
        reusing reades and writers whenever possible. Therefore we must ensure
        that only one reader and writer is generated for each input and output
        signature.
    """
    @staticmethod
    def GatherArgumentTypes(interfaces):
        """
            Gather all invoker argument types for given interfaces.
        """
        dict = {}
        for iface in interfaces.values():
            InvokerArgumentType.AddObjects(dict, iface.methods)
            InvokerArgumentType.AddObjects(dict, iface.signals)
            InvokerArgumentType.AddObjects(dict, iface.properties)

        return OrderedDict(sorted(dict.items()))

    @staticmethod
    def AddObjects(dict, objects):
        for object in objects.values():
            InvokerArgumentType.AddType(dict, "input", object.input)
            InvokerArgumentType.AddType(dict, "output", object.output)

    @staticmethod
    def AddType(dict, type, sbus_signature):
        """
            Add a new argument type to dictionary if possible.
        """
        # We don't generate readers and writers for empty arguments
        if sbus_signature is None or not sbus_signature.arguments:
            return

        # We don't generate readers and writers for custom handlers
        if Invoker.IsCustomHandler(type, sbus_signature):
            return

        # We generate each reader and writer only once
        if sbus_signature.signature in dict:
            return

        dict[sbus_signature.signature] = sbus_signature.arguments


class InvokerKeygen:
    """ Invoker Keygen is a piece of C code that takes care of
        chaining same request into one.

        SBus code generator tries to reduce amount of generated code by
        reusing keygens whenever possible. Therefore we must ensure
        that only one keygen is generated for each signature.
    """
    @staticmethod
    def BuildKey(sbus_member, sbus_signature, Args=None):
        """
            Return dictionary key for given SBUS member and signature or None
            if no keying is supported for this member.
        """
        args = Args if not None else \
            InvokerKeygen.GatherKeyArguments(sbus_member, sbus_signature)

        if args is None:
            return None

        key = sbus_signature.signature if args else "<no-arguments>"

        for idx, arg in args.items():
            key += ',%d' % idx

        return key

    @staticmethod
    def BuildKeygenName(sbus_member, sbus_signature):
        args = InvokerKeygen.GatherKeyArguments(sbus_member, sbus_signature)

        if args is None:
            return "NULL"

        keygen = "_sbus_key_%s" % sbus_signature.signature

        for idx, arg in args.items():
            keygen += '_%d' % idx

        return keygen

    @staticmethod
    def GatherKeyArguments(sbus_member, sbus_signature):
        """
            Gather list of key arguments for an SBus member with given
            signature.

            Return dictionary of <argument-index, argument> sorted by
            its key index or an empty dictionary if no arguments are
            necessary to construct a key.

            Return None for SBus members that do not allow keying.
        """
        keys = {}

        if sbus_signature is not None:
            for idx, arg in enumerate(sbus_signature.arguments.values()):
                if arg.key is not None:
                    keys[idx] = arg

        if not keys and sbus_member.key is None:
            return None

        return OrderedDict(sorted(keys.items(),
                           key=lambda p: p[1].key))

    @staticmethod
    def GatherKeygens(interfaces):
        """
            Gather all keygens needed to implement given interfaces.
        """
        dict = {}
        for iface in interfaces.values():
            for method in iface.methods.values():
                InvokerKeygen.Add(dict, method, method.input)

            for signal in iface.signals.values():
                InvokerKeygen.Add(dict, signal, signal.input)

        return OrderedDict(sorted(dict.items()))

    @staticmethod
    def Add(dict, sbus_member, sbus_signature):
        """
            Add a new keygen to dictionary if possible.
        """
        args = InvokerKeygen.GatherKeyArguments(sbus_member, sbus_signature)

        if args is None:
            return

        key = InvokerKeygen.BuildKey(sbus_member, sbus_signature, Args=args)
        dict[key] = InvokerKeygen.KeygenPair(sbus_signature, args)

    class KeygenPair:
        def __init__(self, sbus_signature, arguments):
            self.signature = sbus_signature.signature
            self.arguments = arguments


class InvokerCaller:
    """ Caller invoker is a piece of C code that takes care of executing
        an outgoing method, signal or property and returning their output.

        SBus code generator tries to reduce amount of generated code by
        reusing invokers whenever possible. Therefore we must ensure that
        invoker for each input and output signature type is generated only
        once.
    """
    @staticmethod
    def GatherMethodInvokers(interfaces, type):
        """
            Gather all required method invokers for given interfaces.
        """
        dict = {}
        for iface in interfaces.values():
            for method in iface.methods.values():
                if not InvokerCaller.IsWanted(iface, method, type):
                    continue

                InvokerCaller.Add(dict, method.input, method.output)

        return OrderedDict(sorted(dict.items()))

    @staticmethod
    def GatherSignalInvokers(interfaces, type):
        """
            Gather all required signal invokers for given interfaces.
        """
        dict = {}
        for iface in interfaces.values():
            for signal in iface.signals.values():
                if not InvokerCaller.IsWanted(iface, signal, type):
                    continue

                InvokerCaller.Add(dict, signal.input, signal.output)

        return OrderedDict(sorted(dict.items()))

    @staticmethod
    def GatherGetInvokers(interfaces, type):
        """
            Gather all required property getters for given interfaces.
        """
        dict = {}
        for iface in interfaces.values():
            for property in iface.properties.values():
                if not InvokerCaller.IsWanted(iface, property, type):
                    continue

                if not property.isReadable():
                    continue

                InvokerCaller.Add(dict, None, property.output)

        return OrderedDict(sorted(dict.items()))

    @staticmethod
    def GatherSetInvokers(interfaces, type):
        """
            Gather all required property setters for given interfaces.
        """
        dict = {}
        for iface in interfaces.values():
            for property in iface.properties.values():
                if not InvokerCaller.IsWanted(iface, property, type):
                    continue

                if not property.isWritable():
                    continue

                InvokerCaller.Add(dict, property.input, None)

        return OrderedDict(sorted(dict.items()))

    @staticmethod
    def Add(dict, input, output):
        """
            Add a new invoker to dictionary if possible.
        """
        invoker = Invoker(input, output)
        key = "in:%s, out:%s" % (invoker.input.invokerSignature,
                                 invoker.output.invokerSignature)
        if key in dict:
            return

        dict[key] = invoker

    @staticmethod
    def IsWantedSync(interface, member):
        names = ["codegen.Caller", "codegen.SyncCaller"]
        # First see if the member has one of these annotations
        if SBus.Annotation.AtleastOneIsSet(names, member.annotations):
            return SBus.Annotation.CheckIfFalse(names, member.annotations)

        return SBus.Annotation.CheckIfFalse(names, interface.annotations)

    @staticmethod
    def IsWantedAsync(interface, member):
        names = ["codegen.Caller", "codegen.AsyncCaller"]

        # First see if the member has one of these annotations
        if SBus.Annotation.AtleastOneIsSet(names, member.annotations):
            return SBus.Annotation.CheckIfFalse(names, member.annotations)

        return SBus.Annotation.CheckIfFalse(names, interface.annotations)

    @staticmethod
    def IsWanted(interface, member, type):
        if type == "sync":
            return InvokerCaller.IsWantedSync(interface, member)
        elif type == "async":
            return InvokerCaller.IsWantedAsync(interface, member)

        wanted_sync = InvokerCaller.IsWantedSync(interface, member)
        wanted_async = InvokerCaller.IsWantedAsync(interface, member)

        return wanted_sync or wanted_async



""" Automation of ldap_extra_attr suite

:requirement: ldap_extra_attrs
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
from __future__ import print_function
import re
import pytest
import os
import time
from sssd.testlib.common.utils import sssdTools
from constants import ds_instance_name
from sssd.testlib.common.ssh2_python import SSHClient


@pytest.mark.usefixtures('setup_sssd', 'create_posix_usersgroups')
@pytest.mark.ldapextraattrs
class TestLdapExtraAttrs(object):
    """
    This is test case class for ldap ldap_extra_attr suite
    """
    @pytest.mark.tier1
    @pytest.mark.converted('test_ldap_extra_attrs.py', 'test_ldap_extra_attrs__filled')
    def test_0001_bz1362023(self, multihost):
        """
        :title: IDM-SSSD-TC: ldap_extra_attrs: SSSD fails to start
          when ldap_user_extra_attrs contains mail
        :id: 260d62d3-91c1-4d42-b783-df031ad34223
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1362023
        """
        services = "nss, pam, ifp"
        ldap_extra_attr = {'ldap_user_extra_attrs':
                           'mail, firstname:givenname, lastname:sn'}
        # stop sssd service
        multihost.client[0].service_sssd('stop')
        tools = sssdTools(multihost.client[0])
        # remove sssd cache
        tools.remove_sss_cache('/var/lib/sss/db')
        sssd_param = {'services': services}
        tools.sssd_conf('sssd', sssd_param)
        domain_params = {'ldap_user_extra_attrs': ldap_extra_attr}
        tools.sssd_conf('domain/%s' % (ds_instance_name), domain_params)
        start = multihost.client[0].service_sssd('start')
        assert start == 0

    @pytest.mark.tier1
    @pytest.mark.converted('test_ldap_extra_attrs.py', 'test_ldap_extra_attrs__filled')
    def test_0002_givenmail(self, multihost):
        """
        :title: IDM-SSSD-TC: ldap_extra_attrs: Verify the entry of option
         value given_email:mail in cache data
        :id: f0fb818e-5706-4444-915d-15e4f5fc6c9e
        """
        tools = sssdTools(multihost.client[0])
        multihost.client[0].service_sssd('stop')
        tools.remove_sss_cache('/var/lib/sss/db')
        section = "sssd"
        sssd_params = {'services': 'nss, pam'}
        tools.sssd_conf(section, sssd_params, action='update')
        ldap_extra_attr = 'given_email:mail'
        domain_params = {'ldap_user_extra_attrs': ldap_extra_attr}
        domain_section = 'domain/%s' % ds_instance_name
        tools.sssd_conf(domain_section, domain_params, action='update')
        start = multihost.client[0].service_sssd('start')
        lkup = 'getent passwd foo1@%s' % ds_instance_name
        lkup_cmd = multihost.client[0].run_command(lkup, raiseonerr=False)
        assert start == 0 and lkup_cmd.returncode == 0
        ldb_cmd = 'ldbsearch -H /var/lib/sss/db/cache_%s.ldb -b cn=users,' \
                  'cn=%s,cn=sysdb' % (ds_instance_name, ds_instance_name)
        cmd = multihost.client[0].run_command(ldb_cmd, raiseonerr=False)
        find = re.compile(r'given_email\:\sfoo1\@example\.test')
        result = find.search(cmd.stdout_text)
        assert result is not None

    @pytest.mark.tier1
    @pytest.mark.converted('test_schema.py', 'test_schema__ldap_extra_attrs_check_ldb')
    def test_0003_checkldb(self, multihost):
        """
        :title: IDM-SSSD-TC: ldap_extra_attrs: Verify recently added
         attribute should be in cache db along with their value
        :id: 7bc84c52-d6d6-4ac0-89e1-128b01d7f8ae
        :customerscenario: True
        """
        tools = sssdTools(multihost.client[0])
        multihost.client[0].service_sssd('stop')
        # remove sssd cache
        tools.remove_sss_cache('/var/lib/sss/db')
        sssd_section = "sssd"
        sssd_params = {'services': 'nss, pam'}
        tools.sssd_conf(sssd_section, sssd_params, action='update')
        ldap_extra_attr = 'ldap_user_extra_attrs = firstname:cn, ' \
                          'lastname:sn, description:gecos, ' \
                          'user`s_home_directoy:homeDirectory, user_id:uid'
        domain_params = {'ldap_user_extra_attrs': ldap_extra_attr}
        domain_section = 'domain/%s' % ds_instance_name
        tools.sssd_conf(domain_section, domain_params, action='update')
        start = multihost.client[0].service_sssd('start')
        lkup = 'getent passwd foo1@%s' % ds_instance_name
        lkup_cmd = multihost.client[0].run_command(lkup, raiseonerr=False)
        assert start == 0 and lkup_cmd.returncode == 0
        ldb_cmd = 'ldbsearch -H /var/lib/sss/db/cache_%s.ldb -b cn=users,' \
                  'cn=%s,cn=sysdb' % (ds_instance_name, ds_instance_name)
        cmd = multihost.client[0].run_command(ldb_cmd, raiseonerr=False)
        checks = ['firstname: foo1', 'lastname: foo1',
                  'description: foo1 User',
                  'user`s_home_directoy: /home/foo1', 'user_id: foo1']
        for str1 in checks:
            find = re.compile(r'%s' % str1)
            result = find.search(cmd.stdout_text)
            assert result is not None

    @pytest.mark.tier1
    @pytest.mark.converted('test_schema.py', 'test_schema__ldap_extra_attrs_negative_cache')
    def test_0004_negativecache(self, multihost):
        """
        :title: IDM-SSSD-TC: ldap_extra_attrs: Check whether, not added
         parameter of user is displaying in cache or not
        :id: 208d78dd-1af3-468c-ab4b-c98b79a412a3
        """
        tools = sssdTools(multihost.client[0])
        multihost.client[0].service_sssd('stop')
        # remove sssd cache
        tools.remove_sss_cache('/var/lib/sss/db')
        sssd_section = "sssd"
        sssd_params = {'services': 'nss, pam'}
        tools.sssd_conf(sssd_section, sssd_params, action='update')
        ldap_extra_attr = 'number:telephonenumber'
        domain_params = {'ldap_user_extra_attrs': ldap_extra_attr}
        domain_section = 'domain/%s' % ds_instance_name
        tools.sssd_conf(domain_section, domain_params, action='update')
        start = multihost.client[0].service_sssd('start')
        lkup = 'getent passwd foo1@%s' % ds_instance_name
        lkup_cmd = multihost.client[0].run_command(lkup, raiseonerr=False)
        assert start == 0 and lkup_cmd.returncode == 0
        ldb_cmd = 'ldbsearch -H /var/lib/sss/db/cache_%s.ldb -b cn=users,' \
                  'cn=%s,cn=sysdb' % (ds_instance_name, ds_instance_name)
        cmd = multihost.client[0].run_command(ldb_cmd, raiseonerr=False)
        find = re.compile(r'number\:telephonenumber')
        result = find.search(cmd.stdout_text)
        assert result is None

    @pytest.mark.tier1
    @pytest.mark.converted('test_schema.py', 'test_schema__ldap_extra_attrs_extra_email')
    def test_0005_ldapextraattrs(self, multihost):
        """
        :title: IDM-SSSD-TC: ldap_extra_attrs: Check sssd should start with
         options ldap_user_email and ldap_user_extra_attrs and check entries in
         cache
        :id: f10bca5c-ead4-426a-b173-12e9f79f01b4
        """
        tools = sssdTools(multihost.client[0])
        multihost.client[0].service_sssd('stop')
        # remove sssd cache
        tools.remove_sss_cache('/var/lib/sss/db')
        sssd_section = "sssd"
        sssd_params = {'services': 'nss, pam, ifp'}
        tools.sssd_conf(sssd_section, sssd_params, action='update')
        ldap_extra_attr = 'email:mail, firstname:cn, lastname:sn'
        domain_params = {'ldap_user_extra_attrs': ldap_extra_attr}
        domain_section = 'domain/%s' % ds_instance_name
        tools.sssd_conf(domain_section, domain_params, action='update')
        domain_params2 = {'ldap_user_email': 'mail'}
        tools.sssd_conf('domain/%s' % (ds_instance_name), domain_params2)
        start = multihost.client[0].service_sssd('start')
        lkup = 'getent passwd foo1@%s' % ds_instance_name
        lkup_cmd = multihost.client[0].run_command(lkup, raiseonerr=False)
        assert start == 0 and lkup_cmd.returncode == 0
        ldb_cmd = 'ldbsearch -H /var/lib/sss/db/cache_%s.ldb -b cn=users,' \
                  'cn=%s,cn=sysdb' % (ds_instance_name, ds_instance_name)
        cmd = multihost.client[0].run_command(ldb_cmd, raiseonerr=False)
        checks = ['mail: foo1@example.test',
                  'email: foo1@example.test',
                  'firstname: foo1',
                  'lastname: foo1']
        for str1 in checks:
            find = re.compile(r'%s' % str1)
            result = find.search(cmd.stdout_text)
            assert result is not None

    @pytest.mark.tier1
    @pytest.mark.converted('test_schema.py', 'test_schema__ldap_extra_attrs_ifp')
    def test_0006_bz1667252(self, multihost):
        """
        :title: ifp: crash when requesting extra attributes
        :id: 617c7909-039c-48a6-ba1a-79ebedea4186
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1667252
        """
        tools = sssdTools(multihost.client[0])
        sssd_params = {'services': 'nss, pam, ifp'}
        tools.sssd_conf('sssd', sssd_params, action='update')
        ifp_params = {'user_attributes': '+test'}
        tools.sssd_conf('ifp', ifp_params)
        domain_section = 'domain/%s' % ds_instance_name
        domain_params = {'ldap_user_extra_attrs': 'test:homeDirectory'}
        tools.sssd_conf(domain_section, domain_params)
        start = multihost.client[0].service_sssd('start')
        assert start == 0
        sssctl_cmd = 'sssctl user-checks foo1@%s' % (ds_instance_name)
        cmd = multihost.client[0].run_command(sssctl_cmd)
        ret = multihost.client[0].service_sssd('status')
        assert cmd.returncode == 0 and ret == 0

    @staticmethod
    @pytest.mark.tier1
    def test_bz847043(multihost, backupsssdconf):
        """
        :title: Thread issue can cause the application to not get
            any identity information bz847043
        :id: a3f5b5ea-9cc6-11ed-98f2-845cf3eff344
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=847043
        :setup:
            1. Clear sssd caches
            2. Compile the provided C test program and run it
        :steps:
            1. Log in via the ssh
            2. With the fixed version, the program should run to completion:
                ./client-hang
                Cancelling thread
                Joining...
                Joined, trying getpwuid_r call
                Never get here
        :expectedresults:
            1. Should succeed
            2. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        client = multihost.client[0]
        tools.sssd_conf("nss", {'filter_groups': 'root',
                                'filter_users': 'root'},
                        action='update')
        tools.sssd_conf("domain/example1",
                        {'use_fully_qualified_names': False},
                        action='update')
        tools.clear_sssd_cache()
        file_location_c = '/script/sssd_client_hang.c'
        client.transport.put_file(os.path.dirname(os.path.abspath(__file__))
                                  + file_location_c,
                                  '/tmp/sssd_client_hang.c')
        client.run_command("touch /tmp/output")
        client.run_command("chown foo1 /tmp/output")
        client.run_command('gcc -lpthread /tmp/sssd_client_hang.c'
                           ' -o /tmp/client-hang')
        client.run_command("chown foo1 /tmp/client-hang")
        ssh = SSHClient(multihost.client[0].sys_hostname, 'foo1', 'Secret123')
        try:
            ssh.connect()
            ssh.execute_command('cd /tmp; ./client-hang > output')
        except Exception:
            pytest.fail("Ssh login failed.")
        time.sleep(2)
        log_str = multihost.client[0].get_file_contents("/tmp/output").decode('utf-8')
        assert "Never get here" in log_str



""" AD-Provider range retrieval

:requirement: range_retrieval
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
"""
import re
import time
import pytest
from sssd.testlib.common.utils import sssdTools


@pytest.mark.usefixtures('fetch_ca_cert')
@pytest.mark.usefixtures('range_retr_mods', 'create_small_grp_usr', 'create_range_aduser_group')
@pytest.mark.tier1_4
@pytest.mark.rangeretrieval
class TestADRangeRetrieval:
    """ Test Cases for AD default output range modified ported from bash
    :setup:
     1. Fetch the certificates from AD on client and Configure client to use them
     2. On AD-server, from default query policy, in lDAPAdminLimits, set value of MaxValRange to 50
     3. Create a group 'smallgrp' with 50 users members on AD-server
     4. On AD-server, create 200 ADusers and 200 ADgroups
     5. On AD-server, create a ADuser and add it to 200 ADgroups as member
     6. On AD-server, create a ADgroup and add 200 ADusers members to it.
    """
    @staticmethod
    def test_0001_grouplookup_large_members(multihost, adjoin):
        """
        :title: with ldap provider lookup group with large number of users
        :id: 74ecb720-e2d3-4c72-b43a-12cf5e6166d6
        :setup:
         1. Configure sssd.conf to use id_provider = ldap, 'ldap_schema = ad' and it's related option
         2. Set ldap_id_mapping to True in domain section of sssd.conf
         3. Restart SSSD service with cleaned cache and logs
        :steps:
          1. Fetch information of rangegroup and assert groupmembers beyond 100 and 200 are fetched)
          2. Assert 'Base attribute of [member]' in sssd_domain.log
          3. Assert 'Parsed range values: [member][50]' in sssd_domain.log
          4. Log in with a AD-user from a group with large number of members
        :expectedresults:
          1. Fetched group information shows members beyond 100 and 200 range
          2. Expected Log is present in domain log
          3. Expected Log is present in domain log
          4. Successful user log in
        :customerscenario: False
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        client.backup_sssd_conf()
        domain_name = client.get_domain_section_name()
        domain_section = f'domain/{domain_name}'
        binddn = f'cn=Administrator,cn=users,{multihost.ad[0].domain_basedn_entry}'
        sssd_params = {
            'debug_level': '0xFFF0',
            'id_provider': 'ldap',
            'use_fully_qualified_names': 'false',
            'ldap_uri': f'ldaps://{multihost.ad[0].sys_hostname}',
            'ldap_id_mapping': 'True',
            'ldap_schema': 'ad',
            'ldap_default_bind_dn': binddn,
            'ldap_default_authtok': f'{multihost.ad[0].ssh_password}',
            'ldap_referrals': 'false',
            'ldap_tls_cacert': '/etc/openldap/certs/ad_cert.pem'
        }
        client.sssd_conf(domain_section, sssd_params)
        client.clear_sssd_cache()
        time.sleep(5)
        cmd1 = multihost.client[0].run_command('getent group rangegroup', raiseonerr=False)
        time.sleep(3)
        log = multihost.client[0].get_file_contents(f'/var/log/sssd/sssd_{domain_name}.log').decode('utf-8')
        patt = re.compile(r'Base.+attribute.+of.+member.*range.0-49.*is.*member', re.IGNORECASE)
        patt1 = re.compile(r'Parsed.+range.*values.*member.*50', re.IGNORECASE)
        login = client.auth_from_client(f'rangeuser010@{domain_name}', 'Secret123') == 3, 'user log in failed'
        client.restore_sssd_conf()
        assert 'rangeuser0100' in cmd1.stdout_text, 'group information missing rangeuser0100'
        assert 'rangeuser0200' in cmd1.stdout_text, 'group information missing rangeuser0200'
        assert patt.search(log), 'Base atribute of [member] log is missing'
        assert patt1.search(log), 'Parsed range values: member] log is missing'
        assert login, 'user log in failed'

    @staticmethod
    def test_0002_ad_provider_search_base_with_filter(multihost, adjoin):
        """
        :title: ad provider search base with filter bz848031
        :id: 928b2f4e-dd1e-416d-abdd-4345be386f5d
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=848031
        :setup:
          1. Configure sssd with 'id_provider = ad' and it's related options
          2. Configure ldap_group-search_base to 'AD_BASEDN??(rangegroup)' in domain_section
        :steps:
          1. Fetch rangegroup membership and assert groupmembers beyond 100 are not fetched
          2. Assert 'Base attribute of [member]' in sssd_domain.log
          4. Assert 'Parsed range values: [member][50]' in sssd_domain.log
          5. Log in with AD-user from a group with large number of members
        :expectedresults:
          1. Fetched group information does not shows members beyond 100
          2. Log is present
          3. Log is present
          4. User Log in successful
        :customerscenario: False
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        client.backup_sssd_conf()
        domain_name = client.get_domain_section_name()
        domain_section = f'domain/{domain_name}'
        multihost.client[0].service_sssd('stop')
        basedn = multihost.ad[0].domain_basedn_entry
        sssd_params = {
            'debug_level': '0xFFF0',
            'use_fully_qualified_names': 'false',
            'id_provider': 'ad',
            'ad_server': f'{multihost.ad[0].sys_hostname}',
            'ad_domain': domain_name,
            'ldap_group_search_base': f'{basedn}??(cn=rangegroup)'
        }
        client.sssd_conf(domain_section, sssd_params)
        client.clear_sssd_cache()
        patt = re.compile(r'Base.attribute.of.*member.*range.*0.49.*is.*member', re.IGNORECASE)
        patt1 = re.compile(r'Parsed.range.values.*member.*50', re.IGNORECASE)
        cmd1 = multihost.client[0].run_command('getent group rangegroup', raiseonerr=False)
        time.sleep(2)
        log_str = multihost.client[0].get_file_contents(f'/var/log/sssd/sssd_{domain_name}.log').decode('utf-8')
        login = client.auth_from_client(f'rangeuser0150@{domain_name}', 'Secret123') == 3
        client.restore_sssd_conf()
        assert 'rangeuser050' in cmd1.stdout_text, 'group missing rangeuser050 as member'
        assert 'rangeuser0100' not in cmd1.stdout_text, 'rangegroup showing rangeuser0100 as member'
        assert 'rangeuser0200' not in cmd1.stdout_text, 'rangegroup showing rangeuser0200 as member'
        assert patt.search(log_str), 'Base atribute of [member] log is missing'
        assert patt1.search(log_str), 'Parsed range values: member] log is missing'
        assert login, 'user login failed'

    @staticmethod
    def test_0003_ad_provider_userlookup_large_numberof_groups(multihost, adjoin):
        """
        :title: ad provider lookup user belonging to large number of groups
        :id: 87843cff-cf70-4537-adb7-d05bd3d0b3c4
        :setup:
         1. Configure id_provider = ad in the domain section
        :steps:
          1. Run 'id rangeuser'
          2. SSSD domain logs should have logs related to "Base attribute of [memberOf;range=0-49]
             is [memberOf]"
          3. SSSD domain logs should have logs for 'Parsed range values: [memberOf][50]'
          4. User 'rangeuser' log in ssh
        :expectedresults:
          1. The 'id Rangeuser' should show the group membership upto 200 groups
          2. Logs is present
          3. Logs is present
          4. Successful authentication
        :customerscenario: False
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain_name = client.get_domain_section_name()
        domain_section = f'domain/{domain_name}'
        multihost.client[0].service_sssd('stop')
        sssd_params = {
            'debug_level': '0xFFF0',
            'id_provider': 'ad',
            'use_fully_qualified_names': 'false',
            'ad_server': f'{multihost.ad[0].sys_hostname}',
            'ad_domain': f'{domain_name}'
        }
        client.sssd_conf(domain_section, sssd_params)
        client.clear_sssd_cache()
        time.sleep(3)
        cmd1 = multihost.client[0].run_command('getent group rangegroup', raiseonerr=False)
        patt = re.compile(r'Base.attribute.of.*member.*range.*0.49.*is.*member', re.IGNORECASE)
        patt1 = re.compile(r'Parsed.range.values.*member.*50', re.IGNORECASE)
        log_str = multihost.client[0].get_file_contents(f'/var/log/sssd/sssd_{domain_name}.log').decode('utf-8')
        login = client.auth_from_client(f'rangeuser010@{domain_name}', 'Secret123') == 3, 'user log in failed'
        assert 'rangeuser0100' in cmd1.stdout_text, 'rangegroup group is not fetching rangeuser0100'
        assert 'rangeuser0200' in cmd1.stdout_text, 'rangegroup group is not fetching rangeuser0200'
        assert login, 'User log in failed'
        assert patt.search(log_str), 'Base atribute of [member] log is missing'
        assert patt1.search(log_str), 'Parsed range values: member] log is missing'

    @staticmethod
    def test_0004_ad_provider_ldap_user_searchbase_with_filter(multihost, adjoin):
        """
        :title: ad provider ldap user search base with filter
        :id: 27100011-c8c6-46ee-b135-8df50537c0fc
        :setup:
          1. Configure ldap_group-search_base to 'AD_BASEDN??(rangegroup)' in domain_section
        :steps:
          1. Fetch information of rangegroup and assert groupmembers beyond 100 and 200 are fetched)
          2. Assert a log related to "Base attribute of [memberOf;range=0-49]" is in domain log
          3. Assert 'Parsed range values: [member][50]' in sssd_domain.log
          4. Log in with AD-user from a group with large number of members
        :expectedresults:
          1. Fetched group information shows members beyond 100 and 200 range
          2. Log is asserted
          3. Log is asserted
          4. Log in successful
        :customerscenario: False
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain_name = client.get_domain_section_name()
        domain_section = f'domain/{domain_name}'
        basedn = multihost.ad[0].domain_basedn_entry
        multihost.client[0].service_sssd('stop')
        sssd_params = {
            'debug_level': '0xFFF0',
            'id_provider': 'ad',
            'use_fully_qualified_names': 'false',
            'ad_server': f'{multihost.ad[0].sys_hostname}',
            'ldap_user_search_base': f'{basedn}??(cn=rangeuser)',
            'ad_domain': domain_name
        }
        client.sssd_conf(domain_section, sssd_params)
        client.clear_sssd_cache()
        cmd1 = multihost.client[0].run_command('id rangeuser', raiseonerr=False)
        patt = re.compile(r'Base.attribute.of.*member.*range.*0.49.*is.*member', re.IGNORECASE)
        patt1 = re.compile(r'Parsed.range.values.*member.*50', re.IGNORECASE)
        log_str = multihost.client[0].get_file_contents(f'/var/log/sssd/sssd_{domain_name}.log').decode('utf-8')
        login = client.auth_from_client(f'rangeuser010@{domain_name}', 'Secret123') == 3, 'user log in failed'
        assert 'rangegroup0100' in cmd1.stdout_text, 'rangegroup group is not fetching rangeuser0100'
        assert 'rangegroup0200' in cmd1.stdout_text, 'rangegroup group is not fetching rangeuser0200'
        assert patt.search(log_str), 'Base atribute of [member] log is missing'
        assert patt1.search(log_str), 'Parsed range values: member] log is missing'
        assert login, 'user login failed'

    @staticmethod
    def test_0005_setting_up_ldap_disable_range_retrieval_to_true(multihost, adjoin):
        """
        :title: Setting up ldap disable range retrieval to true bz928807 bz916997
        :id: 8b5d762f-f198-4a27-b811-df9b3f9f4306
        :bugzilla:
          https://bugzilla.redhat.com/show_bug.cgi?id=928807
          https://bugzilla.redhat.com/show_bug.cgi?id=916997
        :setup:
          1. Set ldap_disable_range_retrieval = true in domain section of sssd.conf
        :steps:
          1. Fetch information of rangegroup and assert user 'rangeuser' is not in it
          2. Fetch group information of smallgrp and confirm it has 50 members
        :expectedresults:
          1. The rangegroup membership does not return rangeuser as it's member
          2. The smallgrp group has 50 members
        :customerscenario: False
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain_name = client.get_domain_section_name()
        domain_section = f'domain/{domain_name}'
        multihost.client[0].service_sssd('stop')
        sssd_params = {
            'debug_level': '0xFFF0',
            'id_provider': 'ad',
            'use_fully_qualified_names': 'false',
            'ad_server': f'{multihost.ad[0].sys_hostname}',
            'ad_domain': f'{domain_name}',
            'ldap_disable_range_retrieval': 'True'
        }
        client.sssd_conf(domain_section, sssd_params)
        client.clear_sssd_cache()
        cmd1 = multihost.client[0].run_command('getent -s sss group rangegroup', raiseonerr=False)
        cmd2 = multihost.client[0].run_command('getent -s sss group smallgrp', raiseonerr=False)
        assert 'rangeuser' not in cmd1.stdout_text, 'rangegroup is having rangeuser as member'
        assert cmd2.stdout_text.count('ad_user') == 50, 'smallgrp group does not contain 50 members'

    @staticmethod
    def test_0006_setting_up_ldap_disable_range_retrieval_to_false(multihost, adjoin):
        """
        :title: Setting up ldap disable range retrieval to false
        :id: 43d9637c-b2ab-47e3-acc7-22abe0fb85ec
        :bugzilla:
          https://bugzilla.redhat.com/show_bug.cgi?id=928807
          https://bugzilla.redhat.com/show_bug.cgi?id=916997
        :setup:
          1. Set ldap_disable_range_retrieval = false in domain section of sssd.conf
          2. Enable ldap_id_mapping in domain section
        :steps:
          1. fetch information of rangegroup and assert user 'rangeuser' is not in it
          2. fetch group information of smallgrp and confirm it has 50 members
        :expectedresults:
          1. The rangegroup membership returns rangeuser as it's member
          2. The smallgrp group has 50 members
        :customerscenario: False
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain_name = client.get_domain_section_name()
        domain_section = f'domain/{domain_name}'
        multihost.client[0].service_sssd('stop')
        sssd_params = {
            'debug_level': '0xFFF0',
            'id_provider': 'ad',
            'use_fully_qualified_names': 'false',
            'ad_server': f'{multihost.ad[0].sys_hostname}',
            'ad_domain': f'{domain_name}',
            'ldap_id_mapping': 'True',
            'ldap_disable_range_retrieval': 'false'
        }
        client.sssd_conf(domain_section, sssd_params)
        client.clear_sssd_cache()
        cmd1 = multihost.client[0].run_command('getent -s sss group rangegroup', raiseonerr=False)
        cmd2 = multihost.client[0].run_command('getent -s sss group smallgrp', raiseonerr=False)
        assert 'rangeuser' in cmd1.stdout_text, 'rangeuser is not the member of rangegroup group'
        assert cmd2.stdout_text.count('ad_user') == 50, 'smallgrp group doesnot contain 50 members'
